def rdk_ssl_config(cluster):
    """Generate rdkafka config keys from cluster's ssl_config"""
    if cluster._ssl_config is None:  # plaintext connections
        return {}
    else:
        # We obtain the ciphers used by the existing BrokerConnections, so the
        # security level for our rdkafka connections is never lower than that
        ciphers = _get_ciphers_from_sockets(cluster.brokers)
        assert ciphers

        conf = cluster._ssl_config
        rdk_conf = {
            "security.protocol": "ssl",
            "ssl.cipher.suites": ciphers,
            "ssl.ca.location": conf.cafile,
            "ssl.certificate.location": conf.certfile,
            "ssl.key.location": conf.keyfile,
            "ssl.key.password": conf.password}
        # Some of the values above may not have been configured:
        return {k: v for k, v in rdk_conf.items() if v is not None}


def _get_ciphers_from_sockets(brokers):
    """Obtain ciphers currently used by pykafka BrokerConnections"""
    ciphers = set()
    for b in brokers.values():
        ciph = b._connection._socket.cipher()
        if ciph is not None:
            ciphers.add(ciph[0])
    return ','.join(ciphers)




"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__all__ = ["KafkaClient"]

import logging

from .cluster import Cluster
from .handlers import ThreadingHandler
try:
    from .handlers import GEventHandler
except ImportError:
    GEventHandler = None


log = logging.getLogger(__name__)


class KafkaClient(object):
    """
    A high-level pythonic client for Kafka

    NOTE: `KafkaClient` holds weak references to `Topic` instances via
    :class:`pykafka.cluster.TopicDict`. To perform operations directly on these topics,
    such as examining their partition lists, client code must hold a strong reference to
    the topics it cares about. If client code doesn't need to examine `Topic` instances
    directly, no strong references are necessary.

    Notes on Zookeeper: Zookeeper is used by kafka and its clients to store several
    types of information, including broker host strings, partition ownerships, and
    depending on your kafka version, consumer offsets. The kafka-console-* tools rely
    on zookeeper to discover brokers - this is why you can't directly specify a broker
    to these tools and are required to give a zookeeper host string. In theory, this
    insulates you as a user of the console tools from having to care about which
    specific brokers in your kafka cluster might be accessible at any given time.

    In pykafka, the paradigm is slightly different, though the above method is also
    supported. When you instantiate a `KafkaClient`, you can specify either `hosts` or
    `zookeeper_hosts`. `hosts` is a comma-separated list of brokers to which to
    connect, and `zookeeper_hosts` is a zookeeper connection string. If you specify
    `zookeeper_hosts`, it overrides `hosts`. Thus you can create a `KafkaClient`
    that is connected to your kafka cluster by providing either a zookeeper or a
    broker connection string.

    As for why the specific components do and don't require knowledge of the zookeeper
    cluster, there are some different reasons. `SimpleConsumer`, since it does not
    perform consumption balancing, does not actually require access to zookeeper at
    all. Since kafka 0.8.2, consumer offset information is stored by the kafka broker
    itself instead of the zookeeper cluster. The `BalancedConsumer`, by contrast,
    requires explicit knowledge of the zookeeper cluster because it performs
    consumption balancing. Zookeeper stores the information about which consumers own
    which partitions and provides a central repository of that information for all
    consumers to read. The `BalancedConsumer` cannot do what it does without direct
    access to zookeeper for this reason. Note that the `ManagedBalancedConsumer`,
    which works with kafka 0.9 and above, removes this dependency on zookeeper from
    the balanced consumption process by storing partition ownership information in
    the kafka broker.

    The `Producer` is allowed to send messages to whatever partitions it wants. In
    pykafka, by default the partition for each message is chosen randomly to provide
    an even distribution of messages across partitions. The producer actually doesn't
    do anything that requires information stored in zookeeper, and since the
    connection to the kafka cluster is handled by the above-mentioned logic in
    `KafkaClient`, it doesn't need the zookeeper host string at all.
    """
    def __init__(self,
                 hosts='127.0.0.1:9092',
                 zookeeper_hosts=None,
                 socket_timeout_ms=30 * 1000,
                 offsets_channel_socket_timeout_ms=10 * 1000,
                 use_greenlets=False,
                 exclude_internal_topics=True,
                 source_address='',
                 ssl_config=None,
                 broker_version='0.9.0'):
        """Create a connection to a Kafka cluster.

        Documentation for source_address can be found at
        https://docs.python.org/2/library/socket.html#socket.create_connection

        :param hosts: Comma-separated list of kafka hosts to which to connect.
            If `ssl_config` is specified, the ports specified here are assumed
            to be SSL ports
        :type hosts: str
        :param zookeeper_hosts: KazooClient-formatted string of ZooKeeper hosts to which
            to connect. If not `None`, this argument takes precedence over `hosts`
        :type zookeeper_hosts: str
        :param socket_timeout_ms: The socket timeout (in milliseconds) for
            network requests
        :type socket_timeout_ms: int
        :param offsets_channel_socket_timeout_ms: The socket timeout (in
            milliseconds) when reading responses for offset commit and
            offset fetch requests.
        :type offsets_channel_socket_timeout_ms: int
        :param use_greenlets: Whether to perform parallel operations on greenlets
            instead of OS threads
        :type use_greenlets: bool
        :param exclude_internal_topics: Whether messages from internal topics
            (specifically, the offsets topic) should be exposed to the consumer.
        :type exclude_internal_topics: bool
        :param source_address: The source address for socket connections
        :type source_address: str `'host:port'`
        :param ssl_config: Config object for SSL connection
        :type ssl_config: :class:`pykafka.connection.SslConfig`
        :param broker_version: The protocol version of the cluster being connected to.
            If this parameter doesn't match the actual broker version, some pykafka
            features may not work properly.
        :type broker_version: str
        """
        self._seed_hosts = zookeeper_hosts if zookeeper_hosts is not None else hosts
        self._source_address = source_address
        self._socket_timeout_ms = socket_timeout_ms
        self._offsets_channel_socket_timeout_ms = offsets_channel_socket_timeout_ms
        if use_greenlets and not GEventHandler:
            raise ImportError('use_greenlets can only be used when gevent is installed.')
        self._handler = GEventHandler() if use_greenlets else ThreadingHandler()
        self.cluster = Cluster(
            hosts,
            self._handler,
            socket_timeout_ms=self._socket_timeout_ms,
            offsets_channel_socket_timeout_ms=self._offsets_channel_socket_timeout_ms,
            exclude_internal_topics=exclude_internal_topics,
            source_address=self._source_address,
            zookeeper_hosts=zookeeper_hosts,
            ssl_config=ssl_config,
            broker_version=broker_version)
        self.brokers = self.cluster.brokers
        self.topics = self.cluster.topics

    def __repr__(self):
        return "<{module}.{name} at {id_} (hosts={hosts})>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self)),
            hosts=self._seed_hosts,
        )

    def update_cluster(self):
        """Update known brokers and topics.

        Updates each Topic and Broker, adding new ones as found,
        with current metadata from the cluster.
        """
        self.cluster.update()



import unittest
import threading
import time
import pytest

from uuid import uuid4

from testinstances.managed_instance import ManagedInstance

from pykafka import KafkaClient, Broker
from pykafka.connection import BrokerConnection
from pykafka.exceptions import SocketDisconnectedError
from pykafka.handlers import ThreadingHandler
from pykafka.utils.compat import itervalues
from pykafka.test.utils import get_cluster, stop_cluster


class TestBrokerConnection(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        if not isinstance(cls.kafka, ManagedInstance):
            pytest.skip("Only test on ManagedInstance (run locally)")
        cls.client = KafkaClient(cls.kafka.brokers)

        # BrokerConnection
        ports = cls.kafka._port_generator(9092)
        cls.dest_port = next(ports)
        cls.src_port = next(ports)
        cls.conn = BrokerConnection('localhost',
                                 cls.dest_port,
                                 cls.client._handler,
                                 buffer_size=1024 * 1024,
                                 source_host='localhost',
                                 source_port=cls.src_port,
                                 ssl_config=None)

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def test_connection_fails_no_broker(self):
        """Fail connection when no broker proc exists"""
        assert self.conn is not None
        with self.assertRaises(SocketDisconnectedError):
            self.conn.connect(3 * 1000)

    def test_retry_connect(self):
        """Should retry until the connection succeeds."""
        def delayed_make_broker():
            time.sleep(2)
            self.kafka._start_broker_proc(self.dest_port)

        def retry_connect_broker():
            self.conn.connect(3 * 1000, attempts=20)

        # start broker after delay to ensure initial failure
        delay_make_broker_thread = threading.Thread(target=delayed_make_broker)
        delay_make_broker_thread.start()

        # connect to broker, using retry
        connect_broker_thread = threading.Thread(target=retry_connect_broker)
        connect_broker_thread.start()

        # give enough time for broker creation to complete
        delay_make_broker_thread.join(5)
        connect_broker_thread.join(6)

        assert self.conn.connected is True



from __future__ import division
"""
Author: Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Cluster"]
import json
import logging
import random
import time
import weakref
from pkg_resources import parse_version

from kazoo.client import KazooClient

from .broker import Broker
from .exceptions import (ERROR_CODES,
                         GroupCoordinatorNotAvailable,
                         NoBrokersAvailableError,
                         SocketDisconnectedError,
                         LeaderNotFoundError,
                         LeaderNotAvailable,
                         UnicodeException)
from .protocol import (GroupCoordinatorRequest, GroupCoordinatorResponse,
                       API_VERSIONS_090, API_VERSIONS_080)
from .topic import Topic
from .utils.compat import iteritems, itervalues, range, get_string

log = logging.getLogger(__name__)


class TopicDict(dict):
    """Lazy dict, which will also attempt to auto-create unknown topics"""

    def __init__(self, cluster, exclude_internal_topics, *args, **kwargs):
        super(TopicDict, self).__init__(*args, **kwargs)
        self._cluster = weakref.ref(cluster)
        self._exclude_internal_topics = exclude_internal_topics

    def values(self):
        return [self[key] for key in self]

    def __getitem__(self, key):
        try:
            key = get_string(key).encode('ascii')
        except UnicodeEncodeError:
            raise UnicodeException("Topic name '{}' contains non-ascii "
                                   "characters".format(key))
        if self._should_exclude_topic(key):
            raise KeyError("You have configured KafkaClient/Cluster to hide "
                           "double-underscored, internal topics")
        topic_ref = super(TopicDict, self).__getitem__(key)
        if topic_ref is not None and topic_ref() is not None:
            return topic_ref()
        else:
            # Topic exists, but needs to be instantiated locally
            for i in range(self._cluster()._max_connection_retries):
                meta = self._cluster()._get_metadata([key])
                try:
                    topic = Topic(self._cluster(), meta.topics[key])
                except LeaderNotFoundError:
                    log.warning("LeaderNotFoundError encountered during Topic creation")
                    if i == self._cluster()._max_connection_retries - 1:
                        raise
                else:
                    self[key] = weakref.ref(topic)
                    return topic

    def __missing__(self, key):
        log.warning('Topic %s not found. Attempting to auto-create.', key)
        self._create_topic(key)

        # Note that __missing__ is called from within dict.__getitem__, so
        # that's what we should be returning (rather than self.__getitem__)
        return super(TopicDict, self).__getitem__(key)

    def _create_topic(self, topic_name):
        """Auto-create a topic.

        Not exposed in the cluster or broker because this is *only*
        auto-creation.  When there's a real API for creating topics,
        with settings and everything, we'll implement that. To expose just
        this now would be disingenuous, since it's features would be hobbled.
        """
        while True:
            # Auto-creating is as simple as issuing a metadata request
            # solely for that topic.  If topic auto-creation is enabled on the
            # broker, the initial response will carry a LeaderNotAvailable
            # error, otherwise it will be an UnknownTopicOrPartition or
            # possibly a RequestTimedOut
            res = self._cluster()._get_metadata(topics=[topic_name])
            err = res.topics[topic_name].err
            if err == LeaderNotAvailable.ERROR_CODE:
                time.sleep(.1)
            elif err == 0:
                log.info('Topic %s successfully auto-created.', topic_name)
                self._cluster().update()
                break
            else:
                raise ERROR_CODES[err](
                    "Failed to auto-create topic '{}'".format(topic_name))

    def _update_topics(self, metadata):
        """Update topics with fresh metadata.

        :param metadata: Metadata for all topics.
        :type metadata: Dict of `{name, metadata}` where `metadata` is
            :class:`pykafka.protocol.TopicMetadata` and `name` is `bytes`.
        """
        # Remove old topics
        removed = set(self.keys()) - set(metadata.keys())
        if len(removed) > 0:
            log.info("Removing %d topics", len(removed))
        for name in removed:
            log.debug("Removing topic '%s'", name)
            super(TopicDict, self).pop(name)

        # Add/update partition information
        if len(metadata) > 0:
            log.info("Discovered %d topics", len(metadata))
        for name, meta in iteritems(metadata):
            if not self._should_exclude_topic(name):
                if name not in self.keys():
                    self[name] = None  # to be instantiated lazily
                    log.debug("Discovered topic '%s'", name)
                else:
                    # avoid instantiating Topic if it isn't already there
                    ref = super(TopicDict, self).__getitem__(name)
                    if ref is not None and ref() is not None:
                        self[name].update(meta)

    def _should_exclude_topic(self, topic_name):
        """Should this topic be excluded from the list shown to the client?"""
        if not self._exclude_internal_topics:
            return False
        return topic_name.startswith(b"__")


class Cluster(object):
    """
    A Cluster is a high-level abstraction of the collection of brokers and
    topics that makes up a real kafka cluster.
    """
    def __init__(self,
                 hosts,
                 handler,
                 socket_timeout_ms=30 * 1000,
                 offsets_channel_socket_timeout_ms=10 * 1000,
                 exclude_internal_topics=True,
                 source_address='',
                 zookeeper_hosts=None,
                 ssl_config=None,
                 broker_version='0.9.0'):
        """Create a new Cluster instance.

        :param hosts: Comma-separated list of kafka hosts to which to connect.
        :type hosts: str
        :param zookeeper_hosts: KazooClient-formatted string of ZooKeeper hosts to which
            to connect. If not `None`, this argument takes precedence over `hosts`
        :type zookeeper_hosts: str
        :param handler: The concurrency handler for network requests.
        :type handler: :class:`pykafka.handlers.Handler`
        :param socket_timeout_ms: The socket timeout (in milliseconds) for
            network requests
        :type socket_timeout_ms: int
        :param offsets_channel_socket_timeout_ms: The socket timeout (in
            milliseconds) when reading responses for offset commit and
            offset fetch requests.
        :type offsets_channel_socket_timeout_ms: int
        :param exclude_internal_topics: Whether messages from internal topics
            (specifically, the offsets topic) should be exposed to consumers.
        :type exclude_internal_topics: bool
        :param source_address: The source address for socket connections
        :type source_address: str `'host:port'`
        :param ssl_config: Config object for SSL connection
        :type ssl_config: :class:`pykafka.connection.SslConfig`
        :param broker_version: The protocol version of the cluster being connected to.
            If this parameter doesn't match the actual broker version, some pykafka
            features may not work properly.
        :type broker_version: str
        """
        self._seed_hosts = zookeeper_hosts if zookeeper_hosts is not None else hosts
        self._socket_timeout_ms = socket_timeout_ms
        self._offsets_channel_socket_timeout_ms = offsets_channel_socket_timeout_ms
        self._handler = handler
        self._brokers = {}
        self._topics = TopicDict(self, exclude_internal_topics)
        self._source_address = source_address
        self._source_host = self._source_address.split(':')[0]
        self._source_port = 0
        self._ssl_config = ssl_config
        self._zookeeper_connect = zookeeper_hosts
        self._max_connection_retries = 3
        self._max_connection_retries_offset_mgr = 8
        self._broker_version = broker_version
        self._api_versions = None
        self.controller_broker = None
        if ':' in self._source_address:
            self._source_port = int(self._source_address.split(':')[1])
        self.fetch_api_versions()
        self.update()

    def __repr__(self):
        return "<{module}.{name} at {id_} (hosts={hosts})>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self)),
            hosts=self._seed_hosts,
        )

    @property
    def brokers(self):
        """The dict of known brokers for this cluster"""
        return self._brokers

    @property
    def topics(self):
        """The dict of known topics for this cluster

        NOTE: This dict is an instance of :class:`pykafka.cluster.TopicDict`, which uses
        weak references and lazy evaluation to avoid instantiating unnecessary
        `pykafka.Topic` objects. Thus, the values displayed when printing `client.topics`
        on a freshly created :class:`pykafka.KafkaClient` will be `None`. This simply
        means that the topic instances have not yet been created, but they will be
        when `__getitem__` is called on the dictionary.
        """
        return self._topics

    @property
    def handler(self):
        """The concurrency handler for network requests"""
        return self._handler

    def _request_metadata(self, broker_connects, topics):
        return self._request_random_broker(broker_connects,
                                           lambda b: b.request_metadata(topics))

    def _request_random_broker(self, broker_connects, req_fn):
        """Make a request to any broker in broker_connects

        Returns the result of the first successful request

        :param broker_connects: The set of brokers to which to attempt to connect
        :type broker_connects: Iterable of two-element sequences of the format
            (broker_host, broker_port)
        :param req_fn: A function accepting a :class:`pykafka.broker.Broker` as its
            sole argument that returns a :class:`pykafka.protocol.Response`. The
            argument to this function will be the each of the brokers discoverable
            via `broker_connects` in turn.
        :type req_fn: function
        """
        for i in range(self._max_connection_retries):
            for host, port in broker_connects:
                try:
                    broker = Broker(-1, host, int(port), self._handler,
                                    self._socket_timeout_ms,
                                    self._offsets_channel_socket_timeout_ms,
                                    buffer_size=1024 * 1024,
                                    source_host=self._source_host,
                                    source_port=self._source_port,
                                    ssl_config=self._ssl_config,
                                    broker_version=self._broker_version,
                                    api_versions=self._api_versions)
                    response = req_fn(broker)
                    if response is not None:
                        return response
                except SocketDisconnectedError:
                    log.error("Socket disconnected during request for "
                              "broker %s:%s. Continuing.", host, port)
                except Exception as e:
                    log.error('Unable to connect to broker %s:%s. Continuing.', host, port)
                    log.exception(e)

    def _get_metadata(self, topics=None):
        """Get fresh cluster metadata from a broker."""
        # Works either on existing brokers or seed_hosts list
        brokers = random.shuffle([b for b in self.brokers.values() if b.connected])
        if brokers:
            for broker in brokers:
                response = broker.request_metadata(topics)
                if response is not None:
                    return response
        else:
            broker_connects = self._get_broker_connection_info()
            metadata = self._request_metadata(broker_connects, topics)
            if metadata is not None:
                return metadata

        # Couldn't connect anywhere. Raise an error.
        raise NoBrokersAvailableError(
            'Unable to connect to a broker to fetch metadata. See logs.')

    def _get_broker_connection_info(self):
        """Get a list of host:port pairs representing possible broker connections

        For use only when self.brokers is not populated (ie at startup)
        """
        broker_connects = []
        if self._zookeeper_connect is not None:
            broker_connects = self._get_brokers_from_zookeeper(self._zookeeper_connect)
        else:
            broker_connects = [
                [broker_str.split(":")[0], broker_str.split(":")[1].split("/")[0]]
                for broker_str in self._seed_hosts.split(',')]
        return broker_connects

    def _get_brokers_from_zookeeper(self, zk_connect):
        """Build a list of broker connection pairs from a ZooKeeper host

        :param zk_connect: The ZooKeeper connect string of the instance to which to
            connect
        :type zk_connect: str
        """
        zookeeper = KazooClient(zk_connect, timeout=self._socket_timeout_ms / 1000)
        try:
            # This math is necessary due to a nested timeout in KazooClient.
            # KazooClient will attempt to retry its connections only until the
            # start() timeout is reached. Each of those retries will timeout as
            # indicated by the KazooClient kwarg. We do a number of timeouts of
            # self._socket_timeout_ms equal to the number of hosts. This provides
            # the same retrying behavior that pykafka uses above when treating
            # this host string as a list of kafka brokers.
            timeout = (len(zk_connect.split(',')) * self._max_connection_retries *
                       self._socket_timeout_ms) / 1000
            zookeeper.start(timeout=timeout)
        except Exception as e:
            log.error('Unable to connect to ZooKeeper instance %s', zk_connect)
            log.exception(e)
            return []
        else:
            try:
                # get a list of connect strings from zookeeper
                brokers_path = "/brokers/ids/"
                broker_ids = zookeeper.get_children(brokers_path)
                broker_connects = []
                for broker_id in broker_ids:
                    broker_json, _ = zookeeper.get("{}{}".format(brokers_path,
                                                                 broker_id))
                    broker_info = json.loads(broker_json.decode("utf-8"))
                    broker_connects.append((broker_info['host'],
                                            broker_info['port']))
                zookeeper.stop()
                self._zookeeper_connect = zk_connect
                return broker_connects
            except Exception as e:
                log.error('Unable to fetch broker info from ZooKeeper')
                log.exception(e)
                return []

    def _update_brokers(self, broker_metadata, controller_id):
        """Update brokers with fresh metadata.

        :param broker_metadata: Metadata for all brokers.
        :type broker_metadata: Dict of `{name: metadata}` where `metadata` is
            :class:`pykafka.protocol.BrokerMetadata` and `name` is str.
        :param controller_id: The ID of the cluster's controller broker, if applicable
        :type controller_id: int
        """
        # FIXME: A cluster with no topics returns no brokers in metadata
        # Remove old brokers
        removed = set(self._brokers.keys()) - set(broker_metadata.keys())
        if len(removed) > 0:
            log.info('Removing %d brokers', len(removed))
        for id_ in removed:
            log.debug('Removing broker %s', self._brokers[id_])
            self._brokers.pop(id_)
        # Add/update current brokers
        if len(broker_metadata) > 0:
            log.info('Discovered %d brokers', len(broker_metadata))
        for id_, meta in iteritems(broker_metadata):
            if id_ not in self._brokers:
                log.debug('Discovered broker id %s: %s:%s', id_, meta.host, meta.port)
                self._brokers[id_] = Broker.from_metadata(
                    meta, self._handler, self._socket_timeout_ms,
                    self._offsets_channel_socket_timeout_ms,
                    buffer_size=1024 * 1024,
                    source_host=self._source_host,
                    source_port=self._source_port,
                    ssl_config=self._ssl_config,
                    broker_version=self._broker_version,
                    api_versions=self._api_versions)
            elif not self._brokers[id_].connected:
                log.info('Reconnecting to broker id %s: %s:%s', id_, meta.host, meta.port)
                try:
                    self._brokers[id_].connect()
                except SocketDisconnectedError:
                    log.info('Failed to re-establish connection with broker id %s: %s:%s',
                             id_, meta.host, meta.port)
            else:
                broker = self._brokers[id_]
                if meta.host == broker.host and meta.port == broker.port:
                    log.info('Broker %s:%s metadata unchanged. Continuing.',
                             broker.host, broker.port)
                    continue  # no changes
                # TODO: Can brokers update? Seems like a problem if so.
                #       Figure out and implement update/disconnect/reconnect if
                #       needed.
                raise Exception('Broker host/port change detected! %s', broker)
        if controller_id is not None:
            if controller_id not in self._brokers:
                raise KeyError("Controller ID {} not present in cluster".format(
                    controller_id))
            self.controller_broker = self._brokers[controller_id]

    def get_managed_group_descriptions(self):
        """Return detailed descriptions of all managed consumer groups on this cluster

        This function only returns descriptions for consumer groups created via the
        Group Management API, which pykafka refers to as :class:`ManagedBalancedConsumer`s
        """
        descriptions = {}
        for broker in itervalues(self.brokers):
            res = broker.describe_groups([group.group_id for group
                                          in itervalues(broker.list_groups().groups)])
            descriptions.update(res.groups)
        return descriptions

    def get_offset_manager(self, consumer_group):
        log.warning("WARNING: Cluster.get_offset_manager is deprecated since pykafka "
                    "2.3.0. Instead, use Cluster.get_group_coordinator.")
        return self.get_group_coordinator(consumer_group)

    def get_group_coordinator(self, consumer_group):
        """Get the broker designated as the group coordinator for this consumer group.

        Based on Step 1 at https://cwiki.apache.org/confluence/display/KAFKA/Committing+and+fetching+consumer+offsets+in+Kafka

        :param consumer_group: The name of the consumer group for which to
            find the offset manager.
        :type consumer_group: str
        """
        log.info("Attempting to discover offset manager for consumer group '%s'",
                 consumer_group)
        max_connection_retries = self._max_connection_retries_offset_mgr
        for i in range(max_connection_retries):
            if i > 0:
                log.debug("Retrying offset manager discovery")
            time.sleep(i * 2)
            for broker in itervalues(self.brokers):

                req = GroupCoordinatorRequest(consumer_group)
                try:
                    future = broker.handler.request(req)
                except AttributeError:
                    log.error("Broker {} not connected during offset manager discovery"
                              .format(broker.id))
                    if i == max_connection_retries - 1:
                        raise
                    continue
                try:
                    res = future.get(GroupCoordinatorResponse)
                except GroupCoordinatorNotAvailable:
                    log.error('Error discovering offset manager - GroupCoordinatorNotAvailable.')
                    if i == max_connection_retries - 1:
                        raise
                except SocketDisconnectedError:
                    log.warning("Socket disconnected during offset manager discovery")
                    if i == max_connection_retries - 1:
                        raise
                    self.update()
                else:
                    coordinator = self.brokers.get(res.coordinator_id, None)
                    if coordinator is None:
                        raise Exception('Coordinator broker with id {id_} not found'.format(id_=res.coordinator_id))
                    log.info("Found coordinator broker with id %s", res.coordinator_id)
                    return coordinator

    def fetch_api_versions(self):
        """Get API version info from an available broker and save it"""
        version = parse_version(self._broker_version)
        if version < parse_version('0.10.0'):
            log.info("Broker version is too old to use automatic API version "
                     "discovery. Falling back to hardcoded versions list.")
            if version >= parse_version('0.9.0'):
                self._api_versions = API_VERSIONS_090
            else:
                self._api_versions = API_VERSIONS_080
            return

        log.info("Requesting API version information")
        broker_connects = self._get_broker_connection_info()
        for i in range(self._max_connection_retries):
            response = self._request_random_broker(broker_connects,
                                                   lambda b: b.fetch_api_versions())
            if not response:
                raise SocketDisconnectedError()
            if response.api_versions:
                self._api_versions = response.api_versions
                log.info("Got api version info: {}".format(self._api_versions))
                return

    def update(self):
        """Update known brokers and topics."""
        for i in range(self._max_connection_retries):
            log.debug("Updating cluster, attempt {}/{}".format(i + 1, self._max_connection_retries))
            metadata = self._get_metadata()
            if len(metadata.brokers) == 0 and len(metadata.topics) == 0:
                log.warning('No broker metadata found. If this is a fresh cluster, '
                            'this may be due to a bug in Kafka. You can force '
                            'broker metadata to be returned by manually creating '
                            'a topic in the cluster. See '
                            'https://issues.apache.org/jira/browse/KAFKA-2154 '
                            'for information.')
            self._update_brokers(metadata.brokers, metadata.controller_id)
            try:
                self._topics._update_topics(metadata.topics)
            except LeaderNotFoundError:
                log.warning("LeaderNotFoundError encountered. This may be "
                            "because one or more partitions have no available replicas.")
                if i == self._max_connection_retries - 1:
                    raise
            else:
                break



from __future__ import division
"""
Author: Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["BalancedConsumer"]
import logging
import socket
import sys
import time
from uuid import uuid4
import weakref

from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeException, NodeExistsError
from kazoo.recipe.watchers import ChildrenWatch
try:
    from kazoo.handlers.gevent import SequentialGeventHandler
except ImportError:
    SequentialGeventHandler = None
from six import reraise

from .common import OffsetType
from .exceptions import (KafkaException, PartitionOwnedError, ConsumerStoppedException,
                         UnicodeException)
from .membershipprotocol import RangeProtocol
from .simpleconsumer import SimpleConsumer
from .utils.compat import range, get_bytes, itervalues, iteritems, get_string
from .utils.error_handlers import valid_int
try:
    from .handlers import GEventHandler
except ImportError:
    GEventHandler = None
try:
    from . import rdkafka
except ImportError:
    rdkafka = False


log = logging.getLogger(__name__)


def _catch_thread_exception(fn):
    """Sets self._worker_exception when fn raises an exception"""
    def wrapped(self, *args, **kwargs):
        try:
            ret = fn(self, *args, **kwargs)
        except Exception:
            self._worker_exception = sys.exc_info()
        else:
            return ret
    return wrapped


class BalancedConsumer(object):
    """
    A self-balancing consumer for Kafka that uses ZooKeeper to communicate
    with other balancing consumers.

    Maintains a single instance of SimpleConsumer, periodically using the
    consumer rebalancing algorithm to reassign partitions to this
    SimpleConsumer.
    """
    def __init__(self,
                 topic,
                 cluster,
                 consumer_group,
                 fetch_message_max_bytes=1024 * 1024,
                 num_consumer_fetchers=1,
                 auto_commit_enable=False,
                 auto_commit_interval_ms=60 * 1000,
                 queued_max_messages=2000,
                 fetch_min_bytes=1,
                 fetch_error_backoff_ms=500,
                 fetch_wait_max_ms=100,
                 offsets_channel_backoff_ms=1000,
                 offsets_commit_max_retries=5,
                 auto_offset_reset=OffsetType.EARLIEST,
                 consumer_timeout_ms=-1,
                 rebalance_max_retries=5,
                 rebalance_backoff_ms=2 * 1000,
                 zookeeper_connection_timeout_ms=6 * 1000,
                 zookeeper_connect=None,
                 zookeeper_hosts='127.0.0.1:2181',
                 zookeeper=None,
                 auto_start=True,
                 reset_offset_on_start=False,
                 post_rebalance_callback=None,
                 use_rdkafka=False,
                 compacted_topic=False,
                 membership_protocol=RangeProtocol,
                 deserializer=None,
                 reset_offset_on_fetch=True):
        """Create a BalancedConsumer instance

        :param topic: The topic this consumer should consume
        :type topic: :class:`pykafka.topic.Topic`
        :param cluster: The cluster to which this consumer should connect
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param consumer_group: The name of the consumer group this consumer
            should join. Consumer group names are namespaced at the cluster level,
            meaning that two consumers consuming different topics with the same group name
            will be treated as part of the same group.
        :type consumer_group: str
        :param fetch_message_max_bytes: The number of bytes of messages to
            attempt to fetch with each fetch request
        :type fetch_message_max_bytes: int
        :param num_consumer_fetchers: The number of workers used to make
            FetchRequests
        :type num_consumer_fetchers: int
        :param auto_commit_enable: If true, periodically commit to kafka the
            offset of messages already returned from consume() calls. Requires that
            `consumer_group` is not `None`.
        :type auto_commit_enable: bool
        :param auto_commit_interval_ms: The frequency (in milliseconds) at which
            the consumer's offsets are committed to kafka. This setting is
            ignored if `auto_commit_enable` is `False`.
        :type auto_commit_interval_ms: int
        :param queued_max_messages: The maximum number of messages buffered for
            consumption in the internal
            :class:`pykafka.simpleconsumer.SimpleConsumer`
        :type queued_max_messages: int
        :param fetch_min_bytes: The minimum amount of data (in bytes) that the
            server should return for a fetch request. If insufficient data is
            available, the request will block until sufficient data is available.
        :type fetch_min_bytes: int
        :param fetch_error_backoff_ms: *UNUSED*.
            See :class:`pykafka.simpleconsumer.SimpleConsumer`.
        :type fetch_error_backoff_ms: int
        :param fetch_wait_max_ms: The maximum amount of time (in milliseconds)
            that the server will block before answering a fetch request if
            there isn't sufficient data to immediately satisfy `fetch_min_bytes`.
        :type fetch_wait_max_ms: int
        :param offsets_channel_backoff_ms: Backoff time to retry failed offset
            commits and fetches.
        :type offsets_channel_backoff_ms: int
        :param offsets_commit_max_retries: The number of times the offset commit
            worker should retry before raising an error.
        :type offsets_commit_max_retries: int
        :param auto_offset_reset: What to do if an offset is out of range. This
            setting indicates how to reset the consumer's internal offset
            counter when an `OffsetOutOfRangeError` is encountered.
        :type auto_offset_reset: :class:`pykafka.common.OffsetType`
        :param consumer_timeout_ms: Amount of time (in milliseconds) the
            consumer may spend without messages available for consumption
            before returning None.
        :type consumer_timeout_ms: int
        :param rebalance_max_retries: The number of times the rebalance should
            retry before raising an error.
        :type rebalance_max_retries: int
        :param rebalance_backoff_ms: Backoff time (in milliseconds) between
            retries during rebalance.
        :type rebalance_backoff_ms: int
        :param zookeeper_connection_timeout_ms: The maximum time (in
            milliseconds) that the consumer waits while establishing a
            connection to zookeeper.
        :type zookeeper_connection_timeout_ms: int
        :param zookeeper_connect: Deprecated::2.7,3.6 Comma-Separated
            (ip1:port1,ip2:port2) strings indicating the zookeeper nodes to which
            to connect.
        :type zookeeper_connect: str
        :param zookeeper_hosts: KazooClient-formatted string of ZooKeeper hosts to which
            to connect.
        :type zookeeper_hosts: str
        :param zookeeper: A KazooClient connected to a Zookeeper instance.
            If provided, `zookeeper_connect` is ignored.
        :type zookeeper: :class:`kazoo.client.KazooClient`
        :param auto_start: Whether the consumer should begin communicating
            with zookeeper after __init__ is complete. If false, communication
            can be started with `start()`.
        :type auto_start: bool
        :param reset_offset_on_start: Whether the consumer should reset its
            internal offset counter to `self._auto_offset_reset` and commit that
            offset immediately upon starting up
        :type reset_offset_on_start: bool
        :param post_rebalance_callback: A function to be called when a rebalance is
            in progress. This function should accept three arguments: the
            :class:`pykafka.balancedconsumer.BalancedConsumer` instance that just
            completed its rebalance, a dict of partitions that it owned before the
            rebalance, and a dict of partitions it owns after the rebalance. These dicts
            map partition ids to the most recently known offsets for those partitions.
            This function can optionally return a dictionary mapping partition ids to
            offsets. If it does, the consumer will reset its offsets to the supplied
            values before continuing consumption.
            Note that the BalancedConsumer is in a poorly defined state at
            the time this callback runs, so that accessing its properties
            (such as `held_offsets` or `partitions`) might yield confusing
            results.  Instead, the callback should really rely on the
            provided partition-id dicts, which are well-defined.
        :type post_rebalance_callback: function
        :param use_rdkafka: Use librdkafka-backed consumer if available
        :type use_rdkafka: bool
        :param compacted_topic: Set to read from a compacted topic. Forces
            consumer to use less stringent message ordering logic because compacted
            topics do not provide offsets in strict incrementing order.
        :type compacted_topic: bool
        :param membership_protocol: The group membership protocol to which this consumer
            should adhere
        :type membership_protocol: :class:`pykafka.membershipprotocol.GroupMembershipProtocol`
        :param deserializer: A function defining how to deserialize messages returned
            from Kafka. A function with the signature d(value, partition_key) that
            returns a tuple of (deserialized_value, deserialized_partition_key). The
            arguments passed to this function are the bytes representations of a
            message's value and partition key, and the returned data should be these
            fields transformed according to the client code's serialization logic.
            See `pykafka.utils.__init__` for stock implemtations.
        :type deserializer: function
        :param reset_offset_on_fetch: Whether to update offsets during fetch_offsets.
               Disable for read-only use cases to prevent side-effects.
        :type reset_offset_on_fetch: bool
        """
        self._cluster = cluster
        try:
            self._consumer_group = get_string(consumer_group).encode('ascii')
        except UnicodeEncodeError:
            raise UnicodeException("Consumer group name '{}' contains non-ascii "
                                   "characters".format(consumer_group))
        self._topic = topic

        self._auto_commit_enable = auto_commit_enable
        self._auto_commit_interval_ms = valid_int(auto_commit_interval_ms)
        self._fetch_message_max_bytes = valid_int(fetch_message_max_bytes)
        self._fetch_min_bytes = valid_int(fetch_min_bytes)
        self._rebalance_max_retries = valid_int(rebalance_max_retries, allow_zero=True)
        self._num_consumer_fetchers = valid_int(num_consumer_fetchers)
        self._queued_max_messages = valid_int(queued_max_messages)
        self._fetch_wait_max_ms = valid_int(fetch_wait_max_ms, allow_zero=True)
        self._rebalance_backoff_ms = valid_int(rebalance_backoff_ms)
        self._consumer_timeout_ms = valid_int(consumer_timeout_ms,
                                              allow_zero=True, allow_negative=True)
        self._offsets_channel_backoff_ms = valid_int(offsets_channel_backoff_ms)
        self._offsets_commit_max_retries = valid_int(offsets_commit_max_retries,
                                                     allow_zero=True)
        self._auto_offset_reset = auto_offset_reset
        self._zookeeper_connect = zookeeper_connect or zookeeper_hosts
        self._zookeeper_connection_timeout_ms = valid_int(zookeeper_connection_timeout_ms,
                                                          allow_zero=True)
        self._reset_offset_on_start = reset_offset_on_start
        self._post_rebalance_callback = post_rebalance_callback
        self._generation_id = -1
        self._running = False
        self._worker_exception = None
        self._is_compacted_topic = compacted_topic
        self._membership_protocol = membership_protocol
        self._deserializer = deserializer
        self._reset_offset_on_fetch = reset_offset_on_fetch

        if not rdkafka and use_rdkafka:
            raise ImportError("use_rdkafka requires rdkafka to be installed")
        if GEventHandler and isinstance(self._cluster.handler, GEventHandler) and use_rdkafka:
            raise ImportError("use_rdkafka cannot be used with gevent")
        self._use_rdkafka = rdkafka and use_rdkafka

        self._rebalancing_lock = cluster.handler.Lock()
        self._rebalancing_in_progress = self._cluster.handler.Event()
        self._internal_consumer_running = self._cluster.handler.Event()
        self._consumer = None
        self._consumer_id = get_bytes("{hostname}:{uuid}".format(
            hostname=socket.gethostname(),
            uuid=uuid4()
        ))
        self._setting_watches = True

        self._topic_path = '/consumers/{group}/owners/{topic}'.format(
            group=get_string(self._consumer_group),
            topic=self._topic.name)
        self._consumer_id_path = '/consumers/{group}/ids'.format(
            group=get_string(self._consumer_group))

        self._zookeeper = None
        self._owns_zookeeper = zookeeper is None
        if zookeeper is not None:
            self._zookeeper = zookeeper
        if auto_start is True:
            self.start()

    def __del__(self):
        log.debug("Finalising {}".format(self))
        if self._running:
            self.stop()

    def __repr__(self):
        return "<{module}.{name} at {id_} (consumer_group={group})>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self)),
            group=get_string(self._consumer_group)
        )

    def _raise_worker_exceptions(self):
        """Raises exceptions encountered on worker threads"""
        if self._worker_exception is not None:
            reraise(*self._worker_exception)

    @property
    def topic(self):
        """The topic this consumer consumes"""
        return self._topic

    @property
    def partitions(self):
        """A list of the partitions that this consumer consumes"""
        return self._consumer.partitions if self._consumer else dict()

    @property
    def _partitions(self):
        """Convenient shorthand for set of partitions internally held"""
        return set(
            [] if self.partitions is None else itervalues(self.partitions))

    @property
    def held_offsets(self):
        """Return a map from partition id to held offset for each partition"""
        if not self._consumer:
            return None
        return self._consumer.held_offsets

    def start(self):
        """Open connections and join a consumer group."""
        try:
            if self._zookeeper is None:
                self._setup_zookeeper(self._zookeeper_connect,
                                      self._zookeeper_connection_timeout_ms)
            self._zookeeper.ensure_path(self._topic_path)
            self._add_self()
            self._running = True
            self._set_watches()
            self._rebalance()
        except Exception:
            log.exception("Stopping consumer in response to error")
            self.stop()

    def stop(self):
        """Close the zookeeper connection and stop consuming.

        This method should be called as part of a graceful shutdown process.
        """
        log.debug("Stopping {}".format(self))
        with self._rebalancing_lock:
            # We acquire the lock in order to prevent a race condition where a
            # rebalance that is already underway might re-register the zk
            # nodes that we remove here
            self._running = False
        if self._consumer is not None:
            self._consumer.stop()
        if self._owns_zookeeper:
            # NB this should always come last, so we do not hand over control
            # of our partitions until consumption has really been halted
            self._zookeeper.stop()
        else:
            self._remove_partitions(self._get_held_partitions())
            try:
                self._zookeeper.delete(self._path_self)
            except NoNodeException:
                pass
        # additionally we'd want to remove watches here, but there are no
        # facilities for that in ChildrenWatch - as a workaround we check
        # self._running in the watcher callbacks (see further down)

    def _setup_zookeeper(self, zookeeper_connect, timeout):
        """Open a connection to a ZooKeeper host.

        :param zookeeper_connect: The 'ip:port' address of the zookeeper node to
            which to connect.
        :type zookeeper_connect: str
        :param timeout: Connection timeout (in milliseconds)
        :type timeout: int
        """
        kazoo_kwargs = {'timeout': timeout / 1000}
        if GEventHandler and isinstance(self._cluster.handler, GEventHandler):
            kazoo_kwargs['handler'] = SequentialGeventHandler()
        self._zookeeper = KazooClient(zookeeper_connect, **kazoo_kwargs)
        self._zookeeper.start()

    def _setup_internal_consumer(self, partitions=None, start=True):
        """Instantiate an internal SimpleConsumer instance"""
        if partitions is None:
            partitions = []
        # Only re-create internal consumer if something changed.
        if partitions != self._partitions:
            cns = self._get_internal_consumer(partitions=list(partitions), start=start)
            if self._post_rebalance_callback is not None:
                old_offsets = (self._consumer.held_offsets
                               if self._consumer else dict())
                new_offsets = cns.held_offsets
                try:
                    reset_offsets = self._post_rebalance_callback(
                        self, old_offsets, new_offsets)
                except Exception:
                    log.exception("post rebalance callback threw an exception")
                    self._worker_exception = sys.exc_info()
                    return False

                if reset_offsets:
                    cns.reset_offsets(partition_offsets=[
                        (cns.partitions[id_], offset) for
                        (id_, offset) in iteritems(reset_offsets)])
            self._consumer = cns
        if self._consumer and self._consumer._running:
            if not self._internal_consumer_running.is_set():
                self._internal_consumer_running.set()
        else:
            if self._internal_consumer_running.is_set():
                self._internal_consumer_running.clear()
        return True

    def _get_internal_consumer(self, partitions=None, start=True):
        """Instantiate a SimpleConsumer for internal use.

        If there is already a SimpleConsumer instance held by this object,
        disable its workers and mark it for garbage collection before
        creating a new one.
        """
        if partitions is None:
            partitions = []
        reset_offset_on_start = self._reset_offset_on_start
        if self._consumer is not None:
            self._consumer.stop()
            # only use this setting for the first call to
            # _get_internal_consumer. subsequent calls should not
            # reset the offsets, since they can happen at any time
            reset_offset_on_start = False
        Cls = (rdkafka.RdKafkaSimpleConsumer
               if self._use_rdkafka else SimpleConsumer)
        cns = Cls(
            self._topic,
            self._cluster,
            consumer_group=self._consumer_group,
            partitions=partitions,
            auto_commit_enable=self._auto_commit_enable,
            auto_commit_interval_ms=self._auto_commit_interval_ms,
            fetch_message_max_bytes=self._fetch_message_max_bytes,
            fetch_min_bytes=self._fetch_min_bytes,
            num_consumer_fetchers=self._num_consumer_fetchers,
            queued_max_messages=self._queued_max_messages,
            fetch_wait_max_ms=self._fetch_wait_max_ms,
            consumer_timeout_ms=self._consumer_timeout_ms,
            offsets_channel_backoff_ms=self._offsets_channel_backoff_ms,
            offsets_commit_max_retries=self._offsets_commit_max_retries,
            auto_offset_reset=self._auto_offset_reset,
            reset_offset_on_start=reset_offset_on_start,
            auto_start=False,
            compacted_topic=self._is_compacted_topic,
            deserializer=self._deserializer,
            reset_offset_on_fetch=self._reset_offset_on_fetch
        )
        cns.consumer_id = self._consumer_id
        cns.generation_id = self._generation_id
        if start:
            cns.start()
        return cns

    def _get_participants(self):
        """Use zookeeper to get the other consumers of this topic.

        :return: A sorted list of the ids of other consumers of this
            consumer's topic
        """
        try:
            consumer_ids = self._zookeeper.get_children(self._consumer_id_path)
        except NoNodeException:
            log.debug("Consumer group doesn't exist. "
                      "No participants to find")
            return []

        participants = []
        for id_ in consumer_ids:
            try:
                topic, stat = self._zookeeper.get("%s/%s" % (self._consumer_id_path, id_))
                if topic == self._topic.name:
                    participants.append(get_bytes(id_))
            except NoNodeException:
                pass  # node disappeared between ``get_children`` and ``get``
        participants = sorted(participants)
        return participants

    def _build_watch_callback(self, fn, proxy):
        """Return a function that's safe to use as a ChildrenWatch callback

        Fixes the issue from https://github.com/Parsely/pykafka/issues/345
        """
        def _callback(children):
            # discover whether the referenced object still exists
            try:
                proxy.__repr__()
            except ReferenceError:
                return False
            return fn(proxy, children)
        return _callback

    def _set_watches(self):
        """Set watches in zookeeper that will trigger rebalances.

        Rebalances should be triggered whenever a broker, topic, or consumer
        znode is changed in zookeeper. This ensures that the balance of the
        consumer group remains up-to-date with the current state of the
        cluster.
        """
        proxy = weakref.proxy(self)
        _brokers_changed = self._build_watch_callback(BalancedConsumer._brokers_changed, proxy)
        _topics_changed = self._build_watch_callback(BalancedConsumer._topics_changed, proxy)
        _consumers_changed = self._build_watch_callback(BalancedConsumer._consumers_changed, proxy)

        self._setting_watches = True
        # Set all our watches and then rebalance
        broker_path = '/brokers/ids'
        try:
            self._broker_watcher = ChildrenWatch(
                self._zookeeper, broker_path,
                _brokers_changed
            )
        except NoNodeException:
            raise Exception(
                'The broker_path "%s" does not exist in your '
                'ZooKeeper cluster -- is your Kafka cluster running?'
                % broker_path)

        self._topics_watcher = ChildrenWatch(
            self._zookeeper,
            '/brokers/topics',
            _topics_changed
        )

        self._consumer_watcher = ChildrenWatch(
            self._zookeeper, self._consumer_id_path,
            _consumers_changed
        )
        self._setting_watches = False

    def _add_self(self):
        """Register this consumer in zookeeper."""
        self._zookeeper.create(
            self._path_self, self._topic.name, ephemeral=True, makepath=True)

    @property
    def _path_self(self):
        """Path where this consumer should be registered in zookeeper"""
        return '{path}/{id_}'.format(
            path=self._consumer_id_path,
            # get_string is necessary to avoid writing literal "b'" to zookeeper
            id_=get_string(self._consumer_id)
        )

    def _update_member_assignment(self):
        """Decide and assign new partitions for this consumer"""
        for i in range(self._rebalance_max_retries):
            try:
                # If retrying, be sure to make sure the
                # partition allocation is correct.
                participants = self._get_participants()
                if self._consumer_id not in participants:
                    # situation that only occurs if our zk session expired
                    self._add_self()
                    participants.append(self._consumer_id)

                new_partitions = self._membership_protocol.decide_partitions(
                    participants, self._topic.partitions, self._consumer_id)
                if not new_partitions:
                    log.warning("No partitions assigned to consumer %s",
                                self._consumer_id)

                # Update zk with any changes:
                # Note that we explicitly fetch our set of held partitions
                # from zk, rather than assuming it will be identical to
                # `self.partitions`.  This covers the (rare) situation
                # where due to an interrupted connection our zk session
                # has expired, in which case we'd hold zero partitions on
                # zk, but `self._partitions` may be outdated and non-empty
                current_zk_parts = self._get_held_partitions()
                self._remove_partitions(current_zk_parts - new_partitions)
                self._add_partitions(new_partitions - current_zk_parts)
                if self._setup_internal_consumer(new_partitions):
                    log.info('Rebalancing Complete.')
                break
            except PartitionOwnedError as ex:
                if i == self._rebalance_max_retries - 1:
                    log.warning('Failed to acquire partition %s after %d retries.',
                                ex.partition, i)
                    raise
                log.info('Unable to acquire partition %s. Retrying', ex.partition)
                self._cluster.handler.sleep(i * (self._rebalance_backoff_ms / 1000))

    def _rebalance(self):
        """Start the rebalancing process for this consumer

        This method is called whenever a zookeeper watch is triggered.
        """
        # This Event is used to notify about rebalance operation to SimpleConsumer's consume().
        if not self._rebalancing_in_progress.is_set():
            self._rebalancing_in_progress.set()

        if self._consumer is not None:
            self.commit_offsets()
        # this is necessary because we can't stop() while the lock is held
        # (it's not an RLock)
        with self._rebalancing_lock:
            if not self._running:
                raise ConsumerStoppedException
            log.info('Rebalancing consumer "%s" for topic "%s".' % (
                self._consumer_id, self._topic.name))
            self._update_member_assignment()

        if self._rebalancing_in_progress.is_set():
            self._rebalancing_in_progress.clear()

    def _path_from_partition(self, p):
        """Given a partition, return its path in zookeeper.

        :type p: :class:`pykafka.partition.Partition`
        """
        return "%s/%s-%s" % (self._topic_path, p.leader.id, p.id)

    def _remove_partitions(self, partitions):
        """Remove partitions from the zookeeper registry for this consumer.

        :param partitions: The partitions to remove.
        :type partitions: Iterable of :class:`pykafka.partition.Partition`
        """
        for p in partitions:
            try:
                # TODO pass zk node version to make sure we still own this node
                self._zookeeper.delete(self._path_from_partition(p))
            except NoNodeException:
                pass

    def _add_partitions(self, partitions):
        """Add partitions to the zookeeper registry for this consumer.

        :param partitions: The partitions to add.
        :type partitions: Iterable of :class:`pykafka.partition.Partition`
        """
        for p in partitions:
            try:
                self._zookeeper.create(
                    self._path_from_partition(p),
                    value=get_bytes(self._consumer_id),
                    ephemeral=True
                )
            except NodeExistsError:
                raise PartitionOwnedError(p)

    def _get_held_partitions(self):
        """Build a set of partitions zookeeper says we own"""
        zk_partition_ids = set()
        all_partitions = self._zookeeper.get_children(self._topic_path)
        for partition_slug in all_partitions:
            try:
                owner_id, stat = self._zookeeper.get(
                    '{path}/{slug}'.format(
                        path=self._topic_path, slug=partition_slug))
                if owner_id == get_bytes(self._consumer_id):
                    zk_partition_ids.add(int(partition_slug.split('-')[1]))
            except NoNodeException:
                pass  # disappeared between ``get_children`` and ``get``
        return set(self._topic.partitions[_id] for _id in zk_partition_ids)

    @_catch_thread_exception
    def _brokers_changed(self, brokers):
        if not self._running:
            return False  # `False` tells ChildrenWatch to disable this watch
        if self._setting_watches:
            return
        log.debug("Rebalance triggered by broker change ({})".format(
            self._consumer_id))
        self._rebalance()

    @_catch_thread_exception
    def _consumers_changed(self, consumers):
        if not self._running:
            return False  # `False` tells ChildrenWatch to disable this watch
        if self._setting_watches:
            return
        log.debug("Rebalance triggered by consumer change ({})".format(
            self._consumer_id))
        self._rebalance()

    @_catch_thread_exception
    def _topics_changed(self, topics):
        if not self._running:
            return False  # `False` tells ChildrenWatch to disable this watch
        if self._setting_watches:
            return
        log.debug("Rebalance triggered by topic change ({})".format(
            self._consumer_id))
        self._rebalance()

    def reset_offsets(self, partition_offsets=None):
        """Reset offsets for the specified partitions

        For each value provided in `partition_offsets`: if the value is an integer,
        immediately reset the partition's internal offset counter to that value. If
        it's a `datetime.datetime` instance or a valid `OffsetType`, issue a
        `ListOffsetRequest` using that timestamp value to discover the latest offset
        in the latest log segment before that timestamp, then set the partition's
        internal counter to that value.

        :param partition_offsets: (`partition`, `timestamp_or_offset`) pairs to
            reset where `partition` is the partition for which to reset the offset
            and `timestamp_or_offset` is EITHER the timestamp before which to find
            a valid offset to set the partition's counter to OR the new offset the
            partition's counter should be set to
        :type partition_offsets: Sequence of tuples of the form
            (:class:`pykafka.partition.Partition`, int OR `datetime.datetime`)
        """
        self._raise_worker_exceptions()
        if not self._consumer:
            raise ConsumerStoppedException("Internal consumer is stopped")
        self._consumer.reset_offsets(partition_offsets=partition_offsets)

    def consume(self, block=True):
        """Get one message from the consumer

        :param block: Whether to block while waiting for a message
        :type block: bool
        """

        def consumer_timed_out():
            """Indicates whether the consumer has received messages recently"""
            if self._consumer_timeout_ms == -1:
                return False
            disp = (time.time() - self._last_message_time) * 1000.0
            return disp > self._consumer_timeout_ms
        message = None
        self._last_message_time = time.time()
        while message is None and not consumer_timed_out():
            if not self._internal_consumer_running.is_set():
                self._cluster.handler.sleep()
                self._raise_worker_exceptions()
                self._internal_consumer_running.wait(self._consumer_timeout_ms / 1000)
            try:
                # acquire the lock to ensure that we don't start trying to consume from
                # a _consumer that might soon be replaced by an in-progress rebalance
                with self._rebalancing_lock:
                    message = self._consumer.consume(block=block, unblock_event=self._rebalancing_in_progress)

                # If Gevent is used, waiting to acquire _rebalancing lock introduces a race condition.
                # This sleep would ensure that the _rebalance method acquires the _rebalancing_lock
                # Issue: https://github.com/Parsely/pykafka/issues/671
                if self._rebalancing_in_progress.is_set():
                    self._cluster.handler.sleep()
            except (ConsumerStoppedException, AttributeError):
                if not self._running:
                    raise ConsumerStoppedException
            if message:
                self._last_message_time = time.time()
            if not block:
                return message
        return message

    def __iter__(self):
        """Yield an infinite stream of messages until the consumer times out"""
        while True:
            message = self.consume(block=True)
            if not message:
                return
            yield message

    def commit_offsets(self, partition_offsets=None):
        """Commit offsets for this consumer's partitions

        Uses the offset commit/fetch API

        :param partition_offsets: (`partition`, `offset`) pairs to
            commit where `partition` is the partition for which to commit the offset
            and `offset` is the offset to commit for the partition. Note that using
            this argument when `auto_commit_enable` is enabled can cause inconsistencies
            in committed offsets. For best results, use *either* this argument *or*
            `auto_commit_enable`.
        :type partition_offsets: Sequence of tuples of the form
            (:class:`pykafka.partition.Partition`, int)
        """
        self._raise_worker_exceptions()
        if not self._consumer:
            raise KafkaException("Cannot commit offsets - consumer not started")
        return self._consumer.commit_offsets(partition_offsets=partition_offsets)



import unittest2

from pykafka import KafkaClient
from pykafka.test.utils import get_cluster, stop_cluster


class TestPartitionInfo(unittest2.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = b'test-data'
        cls.kafka.create_topic(cls.topic_name, 3, 2)
        cls.client = KafkaClient(cls.kafka.brokers)
        topic = cls.client.topics[cls.topic_name]
        cls.producer = topic.get_producer(min_queued_messages=1)
        cls.total_messages = 99
        for i in range(cls.total_messages):
            cls.producer.produce("message {}".format(i).encode())

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def test_can_get_earliest_offset(self):
        partitions = self.client.topics[self.topic_name].partitions
        for partition in partitions.values():
            self.assertEqual(0, partition.earliest_available_offset())

    def test_can_get_latest_offset(self):
        partitions = self.client.topics[self.topic_name].partitions
        for partition in partitions.values():
            self.assertTrue(partition.latest_available_offset() >= 0)

if __name__ == "__main__":
    unittest2.main()



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Partition"]
import datetime as dt
import logging
import time
import weakref

from .common import OffsetType, EPOCH
from .exceptions import LeaderNotFoundError
from .protocol import PartitionOffsetRequest

log = logging.getLogger(__name__)


class Partition(object):
    """
    A Partition is an abstraction over the kafka concept of a partition.
    A kafka partition is a logical division of the logs for a topic. Its
    messages are totally ordered.
    """
    def __init__(self, topic, id_, leader, replicas, isr):
        """Instantiate a new Partition

        :param topic: The topic to which this Partition belongs
        :type topic: :class:`pykafka.topic.Topic`
        :param id_: The identifier for this partition
        :type id_: int
        :param leader: The broker that is currently acting as the leader for
            this partition.
        :type leader: :class:`pykafka.broker.Broker`
        :param replicas: A list of brokers containing this partition's replicas
        :type replicas: Iterable of :class:`pykafka.broker.Broker`
        :param isr: The current set of in-sync replicas for this partition
        :type isr: :class:`pykafka.broker.Broker`
        """
        self._id = id_
        self._leader = leader
        self._replicas = replicas
        self._isr = isr
        self._topic = weakref.ref(topic)

    def __repr__(self):
        return "<{module}.{name} at {id_} (id={my_id})>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self)),
            my_id=self._id,
        )

    def __lt__(self, other):
        return self._id < other._id

    @property
    def id(self):
        """The identifying int for this partition, unique within its topic"""
        return self._id

    @property
    def leader(self):
        """The broker currently acting as leader for this partition"""
        return self._leader

    @property
    def replicas(self):
        """The list of brokers currently holding replicas of this partition"""
        return self._replicas

    @property
    def isr(self):
        """The current list of in-sync replicas for this partition"""
        return self._isr

    @property
    def topic(self):
        """The topic to which this partition belongs"""
        return self._topic()

    def fetch_offset_limit(self, offsets_before, max_offsets=1):
        """Use the Offset API to find a limit of valid offsets
            for this partition.

        :param offsets_before: Return an offset from before
            this timestamp (in milliseconds). Deprecated::2.7,3.6: do not use int
        :type offsets_before: `datetime.datetime` or int
        :param max_offsets: The maximum number of offsets to return
        :type max_offsets: int
        """
        if isinstance(offsets_before, dt.datetime):
            offsets_before = round((offsets_before - EPOCH).total_seconds() * 1000)
        for i in range(self.topic._cluster._max_connection_retries):
            if i > 0:
                log.debug("Retrying offset limit fetch")
            time.sleep(i * 2)
            request = PartitionOffsetRequest(
                self.topic.name, self.id, offsets_before, max_offsets
            )
            res = self._leader.request_offset_limits([request])
            limit = res.topics[self.topic.name][self._id][0]
            if len(limit) > 0:
                return limit

    def latest_available_offset(self):
        """Get the offset of the next message that would be appended to this partition"""
        return self.fetch_offset_limit(OffsetType.LATEST)[0]

    def earliest_available_offset(self):
        """Get the earliest offset for this partition."""
        return self.fetch_offset_limit(OffsetType.EARLIEST)[0]

    def __hash__(self):
        return hash((self.topic, self.id))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other

    def update(self, brokers, metadata):
        """Update this partition with fresh metadata.

        :param brokers: Brokers on which partitions exist
        :type brokers: List of :class:`pykafka.broker.Broker`
        :param metadata: Metadata for the partition
        :type metadata: :class:`pykafka.protocol.PartitionMetadata`
        """
        try:
            # Check leader
            if metadata.leader != self._leader.id:
                log.info('Updating leader for %s from broker %s to broker %s', self,
                         self._leader.id, metadata.leader)
            self._leader = brokers[metadata.leader]
            # Check Replicas
            if sorted(r.id for r in self.replicas) != sorted(metadata.replicas):
                log.info('Updating replicas list for %s', self)
                self._replicas = [brokers[b] for b in metadata.replicas]
            # Check In-Sync-Replicas
            if sorted(i.id for i in self.isr) != sorted(metadata.isr):
                log.info('Updating in sync replicas list for %s', self)
                self._isr = [brokers[b] for b in metadata.isr]
        except KeyError:
            raise LeaderNotFoundError("Replica for partition %s not available. This is "
                                      "probably because none of its replicas are "
                                      "available.", self.id)



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Broker"]
import logging
import time

from .connection import BrokerConnection
from .exceptions import LeaderNotAvailable, SocketDisconnectedError
from .handlers import RequestHandler
from .protocol import (
    FetchRequest, FetchResponse, ListOffsetRequest, ListOffsetResponse, MetadataRequest,
    MetadataResponse, OffsetCommitRequest, OffsetCommitResponse, OffsetFetchRequest,
    OffsetFetchResponse, ProduceResponse, JoinGroupRequest, JoinGroupResponse,
    SyncGroupRequest, SyncGroupResponse, HeartbeatRequest, HeartbeatResponse,
    LeaveGroupRequest, LeaveGroupResponse, ListGroupsRequest, ListGroupsResponse,
    DescribeGroupsRequest, DescribeGroupsResponse, ApiVersionsRequest,
    ApiVersionsResponse, CreateTopicsRequest, CreateTopicsResponse, DeleteTopicsRequest,
    DeleteTopicsResponse)
from .utils.compat import range, iteritems, get_bytes

log = logging.getLogger(__name__)


def _check_handler(fn):
    """Ensures that self._req_handler is not None before calling fn"""
    def wrapped(self, *args, **kwargs):
        if self._req_handler is None:
            raise SocketDisconnectedError
        return fn(self, *args, **kwargs)
    return wrapped


class Broker(object):
    """
    A Broker is an abstraction over a real kafka server instance.
    It is used to perform requests to these servers.
    """
    def __init__(self,
                 id_,
                 host,
                 port,
                 handler,
                 socket_timeout_ms,
                 offsets_channel_socket_timeout_ms,
                 buffer_size=1024 * 1024,
                 source_host='',
                 source_port=0,
                 ssl_config=None,
                 broker_version="0.9.0",
                 api_versions=None):
        """Create a Broker instance.

        :param id_: The id number of this broker
        :type id_: int
        :param host: The host address to which to connect. An IP address or a
            DNS name
        :type host: str
        :param port: The port on which to connect
        :type port: int
        :param handler: A Handler instance that will be used to service requests
            and responses
        :type handler: :class:`pykafka.handlers.Handler`
        :param socket_timeout_ms: The socket timeout for network requests
        :type socket_timeout_ms: int
        :param offsets_channel_socket_timeout_ms: The socket timeout for network
            requests on the offsets channel
        :type offsets_channel_socket_timeout_ms: int
        :param buffer_size: The size (bytes) of the internal buffer used to
            receive network responses
        :type buffer_size: int
        :param source_host: The host portion of the source address for
            socket connections
        :type source_host: str
        :param source_port: The port portion of the source address for
            socket connections
        :type source_port: int
        :param ssl_config: Config object for SSL connection
        :type ssl_config: :class:`pykafka.connection.SslConfig`
        :param broker_version: The protocol version of the cluster being connected to.
            If this parameter doesn't match the actual broker version, some pykafka
            features may not work properly.
        :type broker_version: str
        :param api_versions: A sequence of :class:`pykafka.protocol.ApiVersionsSpec`
            objects indicating the API version compatibility of this broker
        :type api_versions: Iterable of :class:`pykafka.protocol.ApiVersionsSpec`
        """
        self._connection = None
        self._offsets_channel_connection = None
        self._id = int(id_)
        self._host = host
        self._port = port
        self._source_host = source_host
        self._source_port = source_port
        self._ssl_config = ssl_config
        self._handler = handler
        self._req_handler = None
        self._offsets_channel_req_handler = None
        self._socket_timeout_ms = socket_timeout_ms
        self._offsets_channel_socket_timeout_ms = offsets_channel_socket_timeout_ms
        self._buffer_size = buffer_size
        self._req_handlers = {}
        self._broker_version = broker_version
        self._api_versions = api_versions
        try:
            self.connect()
        except SocketDisconnectedError:
            log.warning("Failed to connect to broker at {host}:{port}. Check the "
                        "`listeners` property in server.config."
                        .format(host=self._host, port=self._port))

    def __repr__(self):
        return "<{module}.{name} at {id_} (host={host}, port={port}, id={my_id})>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self)),
            host=self._host,
            port=self._port,
            my_id=self._id
        )

    @classmethod
    def from_metadata(cls,
                      metadata,
                      handler,
                      socket_timeout_ms,
                      offsets_channel_socket_timeout_ms,
                      buffer_size=64 * 1024,
                      source_host='',
                      source_port=0,
                      ssl_config=None,
                      broker_version="0.9.0",
                      api_versions=None):
        """Create a Broker using BrokerMetadata

        :param metadata: Metadata that describes the broker.
        :type metadata: :class:`pykafka.protocol.BrokerMetadata.`
        :param handler: A Handler instance that will be used to service requests
            and responses
        :type handler: :class:`pykafka.handlers.Handler`
        :param socket_timeout_ms: The socket timeout for network requests
        :type socket_timeout_ms: int
        :param offsets_channel_socket_timeout_ms: The socket timeout for network
            requests on the offsets channel
        :type offsets_channel_socket_timeout_ms: int
        :param buffer_size: The size (bytes) of the internal buffer used to
            receive network responses
        :type buffer_size: int
        :param source_host: The host portion of the source address for
            socket connections
        :type source_host: str
        :param source_port: The port portion of the source address for
            socket connections
        :type source_port: int
        :param ssl_config: Config object for SSL connection
        :type ssl_config: :class:`pykafka.connection.SslConfig`
        :param broker_version: The protocol version of the cluster being connected to.
            If this parameter doesn't match the actual broker version, some pykafka
            features may not work properly.
        :type broker_version: str
        :param api_versions: A sequence of :class:`pykafka.protocol.ApiVersionsSpec`
            objects indicating the API version compatibility of this broker
        :type api_versions: Iterable of :class:`pykafka.protocol.ApiVersionsSpec`
        """
        return cls(metadata.id, metadata.host,
                   metadata.port, handler, socket_timeout_ms,
                   offsets_channel_socket_timeout_ms,
                   buffer_size=buffer_size,
                   source_host=source_host,
                   source_port=source_port,
                   ssl_config=ssl_config,
                   broker_version=broker_version,
                   api_versions=api_versions)

    @property
    def connected(self):
        """Returns True if this object's main connection to the Kafka broker
            is active
        """
        return self._connection.connected

    @property
    def offsets_channel_connected(self):
        """Returns True if this object's offsets channel connection to the
            Kafka broker is active
        """
        if self._offsets_channel_connection:
            return self._offsets_channel_connection.connected
        return False

    @property
    def id(self):
        """The broker's ID within the Kafka cluster"""
        return self._id

    @property
    def host(self):
        """The host to which this broker is connected"""
        return self._host

    @property
    def port(self):
        """The port where the broker is available"""
        return self._port

    @property
    def handler(self):
        """The primary :class:`pykafka.handlers.RequestHandler` for this broker

        This handler handles all requests outside of the commit/fetch api
        """
        return self._req_handler

    @property
    def offsets_channel_handler(self):
        """The offset channel :class:`pykafka.handlers.RequestHandler` for this
            broker

        This handler handles all requests that use the commit/fetch api
        """
        return self._offsets_channel_req_handler

    def connect(self, attempts=3):
        """Establish a connection to the broker server.

        Creates a new :class:`pykafka.connection.BrokerConnection` and a new
        :class:`pykafka.handlers.RequestHandler` for this broker
        """
        self._connection = BrokerConnection(self.host, self.port,
                                            self._handler,
                                            buffer_size=self._buffer_size,
                                            source_host=self._source_host,
                                            source_port=self._source_port,
                                            ssl_config=self._ssl_config)
        self._connection.connect(self._socket_timeout_ms, attempts=attempts)
        self._req_handler = RequestHandler(self._handler, self._connection)
        self._req_handler.start()

    def connect_offsets_channel(self, attempts=3):
        """Establish a connection to the Broker for the offsets channel

        Creates a new :class:`pykafka.connection.BrokerConnection` and a new
        :class:`pykafka.handlers.RequestHandler` for this broker's offsets
        channel
        """
        self._offsets_channel_connection = BrokerConnection(
            self.host, self.port, self._handler,
            buffer_size=self._buffer_size,
            source_host=self._source_host, source_port=self._source_port,
            ssl_config=self._ssl_config)
        self._offsets_channel_connection.connect(self._offsets_channel_socket_timeout_ms,
                                                 attempts=attempts)
        self._offsets_channel_req_handler = RequestHandler(
            self._handler, self._offsets_channel_connection
        )
        self._offsets_channel_req_handler.start()

    def _get_unique_req_handler(self, connection_id):
        """Return a RequestHandler instance unique to the given connection_id

        In some applications, for example the Group Membership API, requests running
        in the same process must be interleaved. When both of these requests are
        using the same RequestHandler instance, the requests are queued and the
        interleaving semantics are not upheld. This method behaves identically to
        self._req_handler if there is only one connection_id per KafkaClient.
        If a single KafkaClient needs to use more than one connection_id, this
        method maintains a dictionary of connections unique to those ids.

        :param connection_id: The unique identifier of the connection to return
        :type connection_id: str
        """
        if len(self._req_handlers) == 0:
            self._req_handlers[connection_id] = self._req_handler
        elif connection_id not in self._req_handlers:
            conn = BrokerConnection(
                self.host, self.port, self._handler, buffer_size=self._buffer_size,
                source_host=self._source_host, source_port=self._source_port)
            conn.connect(self._socket_timeout_ms)
            handler = RequestHandler(self._handler, conn)
            handler.start()
            self._req_handlers[connection_id] = handler
        return self._req_handlers[connection_id]

    @_check_handler
    def fetch_messages(self,
                       partition_requests,
                       timeout=30000,
                       min_bytes=1):
        """Fetch messages from a set of partitions.

        :param partition_requests: Requests of messages to fetch.
        :type partition_requests: Iterable of
            :class:`pykafka.protocol.PartitionFetchRequest`
        :param timeout: the maximum amount of time (in milliseconds)
            the server will block before answering the fetch request if there
            isn't sufficient data to immediately satisfy min_bytes
        :type timeout: int
        :param min_bytes: the minimum amount of data (in bytes) the server
            should return. If insufficient data is available the request will
            block for up to `timeout` milliseconds.
        :type min_bytes: int
        """
        request_class = FetchRequest.get_version_impl(self._api_versions)
        response_class = FetchResponse.get_version_impl(self._api_versions)
        future = self._req_handler.request(request_class(
            partition_requests=partition_requests,
            timeout=timeout,
            min_bytes=min_bytes,
            api_version=response_class.API_VERSION
        ))
        # XXX - this call returns even with less than min_bytes of messages?
        return future.get(response_class, broker_version=self._broker_version)

    @_check_handler
    def produce_messages(self, produce_request):
        """Produce messages to a set of partitions.

        :param produce_request: a request object indicating the messages to
            produce
        :type produce_request: :class:`pykafka.protocol.ProduceRequest`
        """
        if produce_request.required_acks == 0:
            self._req_handler.request(produce_request, has_response=False)
        else:
            future = self._req_handler.request(produce_request)
            return future.get(ProduceResponse)

    @_check_handler
    def request_offset_limits(self, partition_requests):
        """Request offset information for a set of topic/partitions

        :param partition_requests: requests specifying the partitions for which
            to fetch offsets
        :type partition_requests: Iterable of
            :class:`pykafka.protocol.PartitionOffsetRequest`
        """
        request_class = ListOffsetRequest.get_version_impl(self._api_versions)
        response_class = ListOffsetResponse.get_version_impl(self._api_versions)
        future = self._req_handler.request(request_class(partition_requests))
        return future.get(response_class)

    @_check_handler
    def request_metadata(self, topics=None):
        """Request cluster metadata

        :param topics: The topic names for which to request metadata
        :type topics: Iterable of `bytes`
        """
        request_class = MetadataRequest.get_version_impl(self._api_versions)
        response_class = MetadataResponse.get_version_impl(self._api_versions)

        max_retries = 3
        for i in range(max_retries):
            if i > 0:
                log.debug("Retrying")
            time.sleep(i)

            try:
                future = self._req_handler.request(request_class(topics=topics))
                response = future.get(response_class)
            except SocketDisconnectedError:
                log.warning("Encountered SocketDisconnectedError while requesting "
                            "metadata from broker %s:%s. Continuing.",
                            self.host, self.port)
                continue

            for name, topic_metadata in iteritems(response.topics):
                if topic_metadata.err == LeaderNotAvailable.ERROR_CODE:
                    log.warning("Leader not available for topic '%s'.", name)
                for pid, partition_metadata in iteritems(topic_metadata.partitions):
                    if partition_metadata.err == LeaderNotAvailable.ERROR_CODE:
                        log.warning("Leader not available for topic '%s' partition %d.",
                                    name, pid)
            return response

    ######################
    #  Commit/Fetch API  #
    ######################

    def commit_consumer_group_offsets(self,
                                      consumer_group,
                                      consumer_group_generation_id,
                                      consumer_id,
                                      preqs):
        """Commit offsets to Kafka using the Offset Commit/Fetch API

        Commit the offsets of all messages consumed so far by this consumer group with the Offset Commit/Fetch API

        Based on Step 2 here https://cwiki.apache.org/confluence/display/KAFKA/Committing+and+fetching+consumer+offsets+in+Kafka

        :param consumer_group: the name of the consumer group for which to
            commit offsets
        :type consumer_group: str
        :param consumer_group_generation_id: The generation ID for this consumer
            group
        :type consumer_group_generation_id: int
        :param consumer_id: The identifier for this consumer group
        :type consumer_id: str
        :param preqs: Requests indicating the partitions for which offsets
            should be committed
        :type preqs: Iterable of :class:`pykafka.protocol.PartitionOffsetCommitRequest`
        """
        if not self.offsets_channel_connected:
            self.connect_offsets_channel()
        req = OffsetCommitRequest(consumer_group,
                                  get_bytes(consumer_group_generation_id),
                                  consumer_id,
                                  partition_requests=preqs)
        return self._offsets_channel_req_handler.request(req).get(OffsetCommitResponse)

    def fetch_consumer_group_offsets(self, consumer_group, preqs):
        """Fetch the offsets stored in Kafka with the Offset Commit/Fetch API

        Based on Step 2 here https://cwiki.apache.org/confluence/display/KAFKA/Committing+and+fetching+consumer+offsets+in+Kafka

        :param consumer_group: the name of the consumer group for which to
            fetch offsets
        :type consumer_group: str
        :param preqs: Requests indicating the partitions for which offsets
            should be fetched
        :type preqs: Iterable of :class:`pykafka.protocol.PartitionOffsetFetchRequest`
        """
        if not self.offsets_channel_connected:
            self.connect_offsets_channel()
        request_class = OffsetFetchRequest.get_version_impl(self._api_versions)
        response_class = OffsetFetchResponse.get_version_impl(self._api_versions)
        req = request_class(consumer_group, partition_requests=preqs)
        return self._offsets_channel_req_handler.request(req).get(response_class)

    ##########################
    #  Group Membership API  #
    ##########################

    def join_group(self,
                   connection_id,
                   consumer_group,
                   member_id,
                   topic_name,
                   membership_protocol):
        """Send a JoinGroupRequest

        :param connection_id: The unique identifier of the connection on which to make
            this request
        :type connection_id: str
        :param consumer_group: The name of the consumer group to join
        :type consumer_group: bytes
        :param member_id: The ID of the consumer joining the group
        :type member_id: bytes
        :param topic_name: The name of the topic to which to connect, used in protocol
            metadata
        :type topic_name: str
        :param membership_protocol: The group membership protocol to which this request
            should adhere
        :type membership_protocol: :class:`pykafka.membershipprotocol.GroupMembershipProtocol`
        """
        handler = self._get_unique_req_handler(connection_id)
        if handler is None:
            raise SocketDisconnectedError
        future = handler.request(JoinGroupRequest(consumer_group, member_id, topic_name,
                                                  membership_protocol))
        self._handler.sleep()
        return future.get(JoinGroupResponse)

    def leave_group(self, connection_id, consumer_group, member_id):
        """Send a LeaveGroupRequest

        :param connection_id: The unique identifier of the connection on which to make
            this request
        :type connection_id: str
        :param consumer_group: The name of the consumer group to leave
        :type consumer_group: bytes
        :param member_id: The ID of the consumer leaving the group
        :type member_id: bytes
        """
        handler = self._get_unique_req_handler(connection_id)
        if handler is None:
            raise SocketDisconnectedError
        future = handler.request(LeaveGroupRequest(consumer_group, member_id))
        return future.get(LeaveGroupResponse)

    def sync_group(self, connection_id, consumer_group, generation_id, member_id, group_assignment):
        """Send a SyncGroupRequest

        :param connection_id: The unique identifier of the connection on which to make
            this request
        :type connection_id: str
        :param consumer_group: The name of the consumer group to which this consumer
            belongs
        :type consumer_group: bytes
        :param generation_id: The current generation for the consumer group
        :type generation_id: int
        :param member_id: The ID of the consumer syncing
        :type member_id: bytes
        :param group_assignment: A sequence of :class:`pykafka.protocol.MemberAssignment`
            instances indicating the partition assignments for each member of the group.
            When `sync_group` is called by a member other than the leader of the group,
            `group_assignment` should be an empty sequence.
        :type group_assignment: iterable of :class:`pykafka.protocol.MemberAssignment`
        """
        handler = self._get_unique_req_handler(connection_id)
        if handler is None:
            raise SocketDisconnectedError
        future = handler.request(
            SyncGroupRequest(consumer_group, generation_id, member_id, group_assignment))
        return future.get(SyncGroupResponse)

    def heartbeat(self, connection_id, consumer_group, generation_id, member_id):
        """Send a HeartbeatRequest

        :param connection_id: The unique identifier of the connection on which to make
            this request
        :type connection_id: str
        :param consumer_group: The name of the consumer group to which this consumer
            belongs
        :type consumer_group: bytes
        :param generation_id: The current generation for the consumer group
        :type generation_id: int
        :param member_id: The ID of the consumer sending this heartbeat
        :type member_id: bytes
        """
        handler = self._get_unique_req_handler(connection_id)
        if handler is None:
            raise SocketDisconnectedError
        future = handler.request(
            HeartbeatRequest(consumer_group, generation_id, member_id))
        self._handler.sleep()
        return future.get(HeartbeatResponse)

    ########################
    #  Administrative API  #
    ########################
    @_check_handler
    def list_groups(self):
        """Send a ListGroupsRequest"""
        future = self._req_handler.request(ListGroupsRequest())
        return future.get(ListGroupsResponse)

    @_check_handler
    def describe_groups(self, group_ids):
        """Send a DescribeGroupsRequest

        :param group_ids: A sequence of group identifiers for which to return descriptions
        :type group_ids: sequence of str
        """
        future = self._req_handler.request(DescribeGroupsRequest(group_ids))
        return future.get(DescribeGroupsResponse)

    ############################
    # Create/Delete Topics API #
    ############################
    @_check_handler
    def create_topics(self, topic_reqs, timeout=0):
        """Create topics via the Topic Creation API

        :param topic_reqs: The topic creation requests to issue
        :type topics: Iterable of :class:`pykafka.protocol.CreateTopicRequest`
        :param timeout: The time in ms to wait for a topic to be completely created.
            Values <= 0 will trigger topic creation and return immediately.
        :type timeout: int
        """
        future = self._req_handler.request(CreateTopicsRequest(topic_reqs,
                                                               timeout=timeout))
        return future.get(CreateTopicsResponse)

    @_check_handler
    def delete_topics(self, topics, timeout=0):
        """Delete topics viaa the Topic Deletion API

        :param topics: The names of the topics to delete
        :type topics: Iterable of str
        :param timeout: The time in ms to wait for a topic to be completely deleted.
            Values <= 0 will trigger topic deletion and return immediately.
        :type timeout: int
        """
        future = self._req_handler.request(DeleteTopicsRequest(topics, timeout=timeout))
        return future.get(DeleteTopicsResponse)

    @_check_handler
    def fetch_api_versions(self):
        """Fetch supported API versions from this broker"""
        future = self._req_handler.request(ApiVersionsRequest())
        return future.get(ApiVersionsResponse.get_version_impl(self._api_versions))



from __future__ import print_function

import argparse
import calendar
import datetime as dt
import sys
import time
from pkg_resources import parse_version

import tabulate

import pykafka
from pykafka.common import OffsetType
from pykafka.protocol import PartitionOffsetCommitRequest, CreateTopicRequest
from pykafka.utils.compat import PY3, iteritems

#
# Helper Functions
#


def fetch_offsets(client, topic, offset):
    """Fetch raw offset data from a topic.

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`pykafka.topic.Topic`
    :param offset: Offset to fetch. Can be earliest, latest or a datetime.
    :type offset: :class:`pykafka.common.OffsetType` or
        :class:`datetime.datetime`
    :returns: {partition_id: :class:`pykafka.protocol.OffsetPartitionResponse`}
    """
    if offset.lower() == 'earliest':
        return topic.earliest_available_offsets()
    elif offset.lower() == 'latest':
        return topic.latest_available_offsets()
    else:
        offset = dt.datetime.strptime(offset, "%Y-%m-%dT%H:%M:%S")
        offset = int(calendar.timegm(offset.utctimetuple()) * 1000)
        return topic.fetch_offset_limits(offset)


def fetch_consumer_lag(client, topic, consumer_group):
    """Get raw lag data for a topic/consumer group.

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`pykafka.topic.Topic`
    :param consumer_group: Name of the consumer group to fetch lag for.
    :type consumer_groups: :class:`str`
    :returns: dict of {partition_id: (latest_offset, consumer_offset)}
    """
    latest_offsets = fetch_offsets(client, topic, 'latest')
    consumer = topic.get_simple_consumer(consumer_group=consumer_group,
                                         auto_start=False, 
                                         reset_offset_on_fetch=False)
    current_offsets = consumer.fetch_offsets()
    return {p_id: (latest_offsets[p_id].offset[0], res.offset)
            for p_id, res in current_offsets}


#
# Commands
#

def consume_topic(client, args):
    """Dump messages from a topic to a file or stdout.

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`str`
    """
    # Don't auto-create topics.
    if args.topic not in client.topics:
        raise ValueError('Topic {} does not exist.'.format(args.topic))
    topic = client.topics[args.topic]
    consumer = topic.get_simple_consumer("pykafka_cli",
                                         consumer_timeout_ms=100,
                                         auto_offset_reset=OffsetType.LATEST,
                                         reset_offset_on_start=True,
                                         auto_commit_enable=True,
                                         auto_commit_interval_ms=3000)
    num_consumed = 0
    while num_consumed < args.limit:
        msg = consumer.consume()
        if not msg or not msg.value:
            continue
        # Coming from Kafka, this is utf-8.
        msg = msg.value.decode('utf-8', errors='replace') + u'\n'
        args.outfile.write(msg if PY3 else msg.encode('utf-8'))
        num_consumed += 1
    args.outfile.close()


def desc_topic(client, args):
    """Print detailed information about a topic.

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`str`
    """
    # Don't auto-create topics.
    if args.topic not in client.topics:
        raise ValueError('Topic {} does not exist.'.format(args.topic))
    topic = client.topics[args.topic]
    print('Topic: {}'.format(topic.name))
    print('Partitions: {}'.format(len(topic.partitions)))
    print('Replicas: {}'.format(len(list(topic.partitions.values())[0].replicas)))
    print(tabulate.tabulate(
        [(p.id, p.leader.id, [r.id for r in p.replicas], [r.id for r in p.isr])
         for p in topic.partitions.values()],
        headers=['Partition', 'Leader', 'Replicas', 'ISR'],
        numalign='center',
    ))


def print_managed_consumer_groups(client, args):
    """Get Kafka-managed consumer groups for a topic.

        :param client: KafkaClient connected to the cluster.
        :type client:  :class:`pykafka.KafkaClient`
        :param topic:  Name of the topic.
        :type topic:  :class:`str`
    """
    if args.topic not in client.topics:
        raise ValueError('Topic {} does not exist.'.format(args.topic))
    consumer_groups = {}
    brokers = client.brokers
    for broker_id, broker in iteritems(brokers):
        groups = broker.list_groups().groups.keys()
        groups_metadata = broker.describe_groups(group_ids=groups).groups
        for group_id, describe_group_response in iteritems(groups_metadata):
            members = describe_group_response.members
            for member_id, member in iteritems(members):
                topics = member.member_metadata.topic_names
                if args.topic in topics:
                    consumer_groups[group_id] = describe_group_response

    print('Topic: {}'.format(args.topic))
    print(tabulate.tabulate(
        [(group_id, x.state, x.protocol, x.protocol_type)
         for group_id, x in iteritems(consumer_groups)],
        headers=['GroupId', 'State', 'Protocol', 'ProtocolType']
    ))


def print_consumer_lag(client, args):
    """Print lag for a topic/consumer group.

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`str`
    :param consumer_group: Name of the consumer group to fetch offsets for.
    :type consumer_groups: :class:`str`
    """
    # Don't auto-create topics.
    if args.topic not in client.topics:
        raise ValueError('Topic {} does not exist.'.format(args.topic))
    topic = client.topics[args.topic]

    lag_info = fetch_consumer_lag(client, topic, args.consumer_group)
    lag_info = [(k, '{:,}'.format(v[0] - v[1]), v[0], v[1])
                for k, v in iteritems(lag_info)]
    print(tabulate.tabulate(
        lag_info,
        headers=['Partition', 'Lag', 'Latest Offset', 'Current Offset'],
        numalign='center',
    ))

    total = sum(int(i[1].replace(',', '')) for i in lag_info)
    print('\n Total lag: {:,} messages.'.format(total))


def print_offsets(client, args):
    """Print offsets for a topic

    NOTE: Time-based offset lookups are not precise, but are based on segment
          boundaries. If there is only one segment, as when Kafka has just
          started, the only offsets found will be [0, <latest_offset>].

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`str`
    :param offset: Offset to reset to. Can be earliest, latest or a datetime.
        Using a datetime will reset the offset to the latest message published
        *before* the datetime.
    :type offset: :class:`pykafka.common.OffsetType` or
        :class:`datetime.datetime`
    """
    # Don't auto-create topics.
    if args.topic not in client.topics:
        raise ValueError('Topic {} does not exist.'.format(args.topic))
    topic = client.topics[args.topic]

    offsets = fetch_offsets(client, topic, args.offset)
    print(tabulate.tabulate(
        [(k, v.offset[0]) for k, v in iteritems(offsets)],
        headers=['Partition', 'Offset'],
        numalign='center',
    ))


def print_topics(client, args):
    """Print all topics in the cluster.

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    """
    print(tabulate.tabulate(
        [(t.name,
          len(t.partitions),
          len(list(t.partitions.values())[0].replicas) - 1)
         for t in client.topics.values()],
        headers=['Topic', 'Partitions', 'Replication'],
        numalign='center',
    ))


def reset_offsets(client, args):
    """Reset offset for a topic/consumer group.

    NOTE: Time-based offset lookups are not precise, but are based on segment
          boundaries. If there is only one segment, as when Kafka has just
          started, the only offsets found will be [0, <latest_offset>].

    :param client: KafkaClient connected to the cluster.
    :type client:  :class:`pykafka.KafkaClient`
    :param topic:  Name of the topic.
    :type topic:  :class:`str`
    :param consumer_group: Name of the consumer group to reset offsets for.
    :type consumer_groups: :class:`str`
    :param offset: Offset to reset to. Can be earliest, latest or a datetime.
        Using a datetime will reset the offset to the latest message published
        *before* the datetime.
    :type offset: :class:`pykafka.common.OffsetType` or
        :class:`datetime.datetime`
    """
    # Don't auto-create topics.
    if args.topic not in client.topics:
        raise ValueError('Topic {} does not exist.'.format(args.topic))
    topic = client.topics[args.topic]

    # Build offset commit requests.
    offsets = fetch_offsets(client, topic, args.offset)
    tmsp = int(time.time() * 1000)
    reqs = [PartitionOffsetCommitRequest(topic.name,
                                         partition_id,
                                         res.offset[0],
                                         tmsp,
                                         b'kafka-tools')
            for partition_id, res in iteritems(offsets)]

    # Send them to the appropriate broker.
    broker = client.cluster.get_group_coordinator(args.consumer_group)
    broker.commit_consumer_group_offsets(
        args.consumer_group, 1, b'kafka-tools', reqs
    )


def create_topic(client, args):
    if parse_version(args.broker_version) < parse_version('0.10.1'):
        raise ValueError("The topic creation API is not usable on brokers older than "
                         "0.10.1. Use --broker_version to specify the version")
    topic_req = CreateTopicRequest(args.topic, args.num_partitions,
                                   args.replication_factor,
                                   args.replica_assignment or [],
                                   args.config_entries or [])
    client.cluster.controller_broker.create_topics([topic_req], args.timeout)


def delete_topic(client, args):
    if parse_version(args.broker_version) < parse_version('0.10.1'):
        raise ValueError("The topic deletion API is not usable on brokers older than "
                         "0.10.1. Use --broker_version to specify the version")
    client.cluster.controller_broker.delete_topics([args.topic], args.timeout)


def _encode_utf8(string):
    """Converts argument to UTF-8-encoded bytes.

    Used to make dict lookups work in Python 3 because topics and things are
    bytes.
    """
    return string.encode('utf-8', errors='replace')


def _add_consumer_group(parser):
    """Add consumer_group to arg parser."""
    parser.add_argument('consumer_group',
                        metavar='CONSUMER_GROUP',
                        help='Consumer group name.',
                        type=_encode_utf8)


def _add_limit(parser):
    """Add limit option to arg parser."""
    parser.add_argument('-l', '--limit',
                        help='Number of messages to consume '
                             '(default: %(default)s)',
                        type=int, default=10)


def _add_timeout(parser):
    parser.add_argument('-t', '--timeout',
                        help='Time in ms to wait for the operation to complete'
                             '(default: %(default)s)',
                        type=int, default=5000)


def _add_offset(parser):
    """Add offset to arg parser."""
    parser.add_argument('offset',
                        metavar='OFFSET',
                        type=str,
                        help='Offset to fetch. Can be EARLIEST, LATEST, or a '
                             'datetime in the format YYYY-MM-DDTHH:MM:SS.')


def _add_outfile(parser):
    """ Add outfile option to arg parser."""
    parser.add_argument('-o', '--outfile',
                        help='output file (defaults to stdout)',
                        type=argparse.FileType('w'), default=sys.stdout)


def _add_topic(parser):
    """Add topic to arg parser."""
    parser.add_argument('topic',
                        metavar='TOPIC',
                        help='Topic name.',
                        type=_encode_utf8)


def _get_arg_parser():
    output = argparse.ArgumentParser(description='Tools for Kafka.')

    # Common arguments
    output.add_argument('-b', '--broker',
                        required=False,
                        default='localhost:9092',
                        dest='host',
                        help='host:port of any Kafka broker. '
                             '[default: localhost:9092]')
    output.add_argument('-o', '--broker_version',
                        required=False,
                        default='0.9.0',
                        dest='broker_version',
                        help="The version string of the broker with which to communicate")

    subparsers = output.add_subparsers(help='Commands', dest='command')

    # Consume to File
    parser = subparsers.add_parser(
        'consume_topic',
        help='Dump messages for a topic to a file or stdout.')
    parser.set_defaults(func=consume_topic)
    _add_topic(parser)
    _add_limit(parser)
    _add_outfile(parser)

    # Desc Topic
    parser = subparsers.add_parser(
        'desc_topic',
        help='Print detailed info for a topic.'
    )
    parser.set_defaults(func=desc_topic)
    _add_topic(parser)

    # Get consumer groups for a topic
    parser = subparsers.add_parser(
        'print_managed_consumer_groups',
        help='Get consumer groups for a topic'
    )
    parser.set_defaults(func=print_managed_consumer_groups)
    _add_topic(parser)

    # Print Consumer Lag
    parser = subparsers.add_parser(
        'print_consumer_lag',
        help='Get consumer lag for a topic.'
    )
    parser.set_defaults(func=print_consumer_lag)
    _add_topic(parser)
    _add_consumer_group(parser)

    # Print Broker Offsets
    parser = subparsers.add_parser(
        'print_offsets',
        help='Fetch broker offsets for a topic'
    )
    parser.set_defaults(func=print_offsets)
    _add_topic(parser)
    _add_offset(parser)

    # Print Topics
    parser = subparsers.add_parser(
        'print_topics',
        help='Print information about all topics in the cluster.'
    )
    parser.set_defaults(func=print_topics)

    # Reset Offsets
    parser = subparsers.add_parser(
        'reset_offsets',
        help='Reset offsets for a topic/consumer group'
    )
    parser.set_defaults(func=reset_offsets)
    _add_topic(parser)
    _add_consumer_group(parser)
    _add_offset(parser)

    parser = subparsers.add_parser(
        'create_topic',
        help='Create a topic'
    )
    parser.set_defaults(func=create_topic)
    _add_topic(parser)
    _add_timeout(parser)
    parser.add_argument('-p', '--num_partitions',
                        help='Number of partitions to be created. -1 indicates unset. '
                             '(default: %(default)s)',
                        type=int, default=1)
    parser.add_argument('-r', '--replication_factor',
                        help='Replication factor for the topic. -1 indicates unset. '
                             '(default: %(default)s)',
                        type=int, default=1)
    parser.add_argument('-a', '--replica_assignment',
                        help='Replica assignment among kafka brokers for this topic '
                             'partitions. If this is set num_partitions and '
                             'replication_factor must be unset. Represent as a JSON '
                             'object with partition IDs as keys as lists of node IDs '
                             'as values',
                        type=_encode_utf8)
    parser.add_argument('-c', '--config_entries',
                        help='Topic level configuration for topic to be set. Represent '
                             'as a JSON object.',
                        type=_encode_utf8)

    parser = subparsers.add_parser(
        'delete_topic',
        help='Delete a topic'
    )
    parser.set_defaults(func=delete_topic)
    _add_topic(parser)
    _add_timeout(parser)

    return output


def main():
    parser = _get_arg_parser()
    args = parser.parse_args()
    if args.command:
        client = pykafka.KafkaClient(hosts=args.host, broker_version=args.broker_version)
        args.func(client, args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["unpack_from"]
import itertools
import struct
from .compat import range


def unpack_from(fmt, buff, offset=0):
    """A customized version of `struct.unpack_from`

    This is a conveinence function that makes decoding the arrays,
    strings, and byte arrays that we get from Kafka significantly
    easier. It takes the same arguments as `struct.unpack_from` but
    adds 3 new formats:

    * Wrap a section in `[]` to indicate an array. e.g.: `[ii]`
    * `S` for strings (int16 followed by byte array)
    * `Y` for byte arrays (int32 followed by byte array)

    Spaces are ignored in the format string, allowing more readable formats

    NOTE: This may be a performance bottleneck. We're avoiding a lot of memory
          allocations by using the same buffer, but if we could call
          `struct.unpack_from` only once, that's about an order of magnitude
          faster. However, constructing the format string to do so would erase
          any gains we got from having the single call.
    """
    fmt = fmt.replace(' ', '')
    if fmt[0] in '!><':
        fmt = fmt[1:]  # It's always network ordering

    output = _unpack(fmt, buff, offset, 1)[0]

    # whole-message arrays come back weird
    if fmt[0] == '[' and len(output) == 1:
        output = output[0]

    return output


def _unpack(fmt, buff, offset, count=1):
    """Recursive call for unpacking

    :param fmt: The struct format string
    :type fmt: str
    :param buff: The buffer into which to unpack
    :type buff: buffer
    :param offset: The offset at which to start unpacking
    :type offset: int
    :param count: The number of items in the array
    :type count: int
    """
    items = []
    array_fmt = None
    for i, ch in enumerate(fmt):
        if array_fmt is not None:
            if ch == ']':
                if array_fmt.count('[') == array_fmt.count(']'):
                    # array format done, call _unpack for this format string
                    count = struct.unpack_from('!i', buff, offset)[0]
                    array_item, offset = _unpack_array(array_fmt, buff,
                                                       offset + 4, count)
                    items.append(array_item)
                    array_fmt = None
                    continue  # done with this
            # not done yet, append to ongoing format
            array_fmt += ch
        elif ch == '[':
            array_fmt = ''  # starts building string for array unpack
        else:
            if ch in 'SY':
                len_fmt = '!h' if ch == 'S' else '!i'
                len_ = struct.unpack_from(len_fmt, buff, offset)[0]
                offset += struct.calcsize(len_fmt)
                if len_ == -1:
                    items.append(None)
                    continue
                ch = '%ds' % len_
            items.extend(struct.unpack_from('!' + ch, buff, offset))
            offset += struct.calcsize(ch)
    return tuple(items), offset


def _unpack_array(fmt, buff, offset, count):
    """Unpack an array of items.

    :param fmt: The struct format string
    :type fmt: str
    :param buff: The buffer into which to unpack
    :type buff: buffer
    :param offset: The offset at which to start unpacking
    :type offset: int
    :param count: The number of items in the array
    :type count: int
    """
    output = []
    for i in range(count):
        item, offset = _unpack(fmt, buff, offset)
        output.append(item)
    if len(fmt) == 1:
        output = list(itertools.chain.from_iterable(output))
    return output, offset



# - coding: utf-8 -
import itertools
import struct
from collections import namedtuple, defaultdict

from .base import Request, Response
from .message import MessageSet
from ..common import CompressionType
from ..utils import struct_helpers
from ..utils.compat import iteritems, itervalues


class ProduceRequest(Request):
    """Produce Request
    Specification::
        ProduceRequest => RequiredAcks Timeout [TopicName [Partition MessageSetSize MessageSet]]
          RequiredAcks => int16
          Timeout => int32
          Partition => int32
          MessageSetSize => int32
    """
    API_KEY = 0

    def __init__(self,
                 compression_type=CompressionType.NONE,
                 required_acks=1,
                 timeout=10000,
                 broker_version='0.9.0'):
        """Create a new ProduceRequest
        ``required_acks`` determines how many acknowledgement the server waits
        for before returning. This is useful for ensuring the replication factor
        of published messages. The behavior is::
            -1: Block until all servers acknowledge
            0: No waiting -- server doesn't even respond to the Produce request
            1: Wait for this server to write to the local log and then return
            2+: Wait for N servers to acknowledge
        :param partition_requests: Iterable of
            :class:`kafka.pykafka.protocol.PartitionProduceRequest` for this request
        :param compression_type: Compression to use for messages
        :param required_acks: see docstring
        :param timeout: timeout (in ms) to wait for the required acks
        """
        # {topic_name: {partition_id: MessageSet}}
        self.msets = defaultdict(
            lambda: defaultdict(
                lambda: MessageSet(compression_type=compression_type,
                                   broker_version=broker_version)
            ))
        self.required_acks = required_acks
        self.timeout = timeout
        self._message_count = 0  # this optimization is not premature

    def __len__(self):
        """Length of the serialized message, in bytes"""
        size = self.HEADER_LEN + 2 + 4 + 4  # acks + timeout + len(topics)
        for topic, parts in iteritems(self.msets):
            # topic name
            size += 2 + len(topic) + 4  # topic name + len(parts)
            # partition + mset size + len(mset)
            size += sum(4 + 4 + len(mset) for mset in itervalues(parts))
        return size

    @property
    def messages(self):
        """Iterable of all messages in the Request"""
        return itertools.chain.from_iterable(
            mset.messages
            for topic, partitions in iteritems(self.msets)
            for partition_id, mset in iteritems(partitions)
        )

    def add_message(self, message, topic_name, partition_id):
        """Add a list of :class:`kafka.common.Message` to the waiting request
        :param messages: an iterable of :class:`kafka.common.Message` to add
        :param topic_name: the name of the topic to publish to
        :param partition_id: the partition to publish to
        """
        self.msets[topic_name][partition_id].messages.append(message)
        self._message_count += 1

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        struct.pack_into('!hii', output, offset,
                         self.required_acks, self.timeout, len(self.msets))
        offset += 10
        for topic_name, partitions in iteritems(self.msets):
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name),
                             topic_name, len(partitions))
            offset += struct.calcsize(fmt)
            for partition_id, message_set in iteritems(partitions):
                mset_len = len(message_set)
                struct.pack_into('!ii', output, offset, partition_id, mset_len)
                offset += 8
                message_set.pack_into(output, offset)
                offset += mset_len
        return output

    def message_count(self):
        """Get the number of messages across all MessageSets in the request."""
        return self._message_count


ProducePartitionResponse = namedtuple(
    'ProducePartitionResponse',
    ['err', 'offset']
)


class ProduceResponse(Response):
    """Produce Response. Checks to make sure everything went okay.
    Specification::
        ProduceResponse => [TopicName [Partition ErrorCode Offset]]
          TopicName => string
          Partition => int32
          ErrorCode => int16
          Offset => int64
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        # TODO: Handle having produced to a non-existent topic (in client)
        fmt = '[S [ihq] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        self.topics = {}
        for (topic, partitions) in response:
            self.topics[topic] = {}
            for partition in partitions:
                pres = ProducePartitionResponse(partition[1], partition[2])
                self.topics[topic][partition[0]] = pres



# - coding: utf-8 -
import struct
from collections import namedtuple

from ..exceptions import ERROR_CODES
from ..utils import Serializable, ApiVersionAware, struct_helpers


class Request(Serializable, ApiVersionAware):
    """Base class for all Requests. Handles writing header information"""
    HEADER_LEN = 21  # constant for all messages
    CLIENT_ID = b'pykafka'
    API_KEY = -1

    @classmethod
    def get_versions(cls):
        return {}

    def _write_header(self, buff, api_version=0, correlation_id=0):
        """Write the header for an outgoing message.

        :param buff: The buffer into which to write the header
        :type buff: buffer
        :param api_version: The "kafka api version id", used for feature flagging
        :type api_version: int
        :param correlation_id: This is a user-supplied integer. It will be
            passed back in the response by the server, unmodified. It is useful
            for matching request and response between the client and server.
        :type correlation_id: int
        """
        fmt = '!ihhih%ds' % len(self.CLIENT_ID)
        struct.pack_into(fmt, buff, 0,
                         len(buff) - 4,  # msglen excludes this int
                         self.API_KEY,
                         api_version,
                         correlation_id,
                         len(self.CLIENT_ID),
                         self.CLIENT_ID)

    def get_bytes(self):
        """Serialize the message

        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        raise NotImplementedError()


class Response(ApiVersionAware):
    """Base class for Response objects."""
    API_KEY = -1

    @classmethod
    def get_versions(cls):
        return {}

    def raise_error(self, err_code, response):
        """Raise an error based on the Kafka error code

        :param err_code: The error code from Kafka
        :param response: The unpacked raw data from the response
        """
        clsname = str(self.__class__).split('.')[-1].split("'")[0]
        raise ERROR_CODES[err_code](
            'Response Type: "%s"\tResponse: %s' % (
                clsname, response))


class MemberAssignment(object):
    """
    Protocol specification::

    MemberAssignment => Version PartitionAssignment
        Version => int16
        PartitionAssignment => [Topic [Partition]]
            Topic => string
            Partition => int32
        UserData => bytes
    """
    def __init__(self, partition_assignment, version=1):
        self.version = version
        self.partition_assignment = partition_assignment

    @classmethod
    def from_bytestring(cls, buff):
        if len(buff) == 0:
            return cls(tuple())
        fmt = 'h [S [i ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        version = response[0]
        partition_assignment = response[1]
        return cls(partition_assignment, version=version)

    def __len__(self):
        # version + len(partition assignment)
        size = 2 + 4
        for topic_name, partitions in self.partition_assignment:
            # len(topic_name) + topic_name + len(partitions)
            size += 2 + len(topic_name) + 4
            size += 4 * len(partitions)
        return size

    def get_bytes(self):
        output = bytearray(len(self))
        offset = 0
        fmt = '!hi'
        struct.pack_into(fmt, output, offset, self.version,
                         len(self.partition_assignment))
        offset += struct.calcsize(fmt)
        for topic_name, partitions in self.partition_assignment:
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name), topic_name,
                             len(partitions))
            offset += struct.calcsize(fmt)
            for partition_id in partitions:
                fmt = '!i'
                struct.pack_into(fmt, output, offset, partition_id)
                offset += struct.calcsize(fmt)
        return output


class ConsumerGroupProtocolMetadata(object):
    """
    Protocol specification::

    ProtocolMetadata => Version Subscription UserData
        Version => int16
        Subscription => [Topic]
            Topic => string
        UserData => bytes
    """
    def __init__(self, version=0, topic_names=None, user_data=b"testuserdata"):
        self.version = version
        self.topic_names = topic_names or [b"dummytopic"]
        self.user_data = user_data

    def __len__(self):
        # version + len(topic names)
        size = 2 + 4
        for topic_name in self.topic_names:
            # len(topic_name) + topic_name
            size += 2 + len(topic_name)
        # len(user data) + user data
        size += 4 + len(self.user_data)
        return size

    def get_bytes(self):
        output = bytearray(len(self))
        offset = 0
        fmt = '!hi'
        struct.pack_into(fmt, output, offset, self.version, len(self.topic_names))
        offset += struct.calcsize(fmt)
        for topic_name in self.topic_names:
            fmt = '!h%ds' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name), topic_name)
            offset += struct.calcsize(fmt)
        fmt = '!i%ds' % len(self.user_data)
        struct.pack_into(fmt, output, offset, len(self.user_data), self.user_data)
        offset += struct.calcsize(fmt)
        return output

    @classmethod
    def from_bytestring(cls, buff):
        if len(buff) == 0:
            return cls()
        fmt = 'h [S] Y'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        version = response[0]
        topic_names = response[1]
        user_data = response[2]
        return cls(version, topic_names, user_data)


ApiVersionsSpec = namedtuple('ApiVersionsSpec', ['key', 'min', 'max'])


# Hardcoded API version specifiers for brokers that don't support ApiVersionsRequest
API_VERSIONS_080 = {
    0: ApiVersionsSpec(0, 0, 0),
    1: ApiVersionsSpec(1, 0, 0),
    2: ApiVersionsSpec(2, 0, 0),
    3: ApiVersionsSpec(3, 0, 0),
    4: ApiVersionsSpec(4, 0, 0),
    5: ApiVersionsSpec(5, 0, 0),
    6: ApiVersionsSpec(6, 0, 0),
    7: ApiVersionsSpec(7, 0, 0),
    8: ApiVersionsSpec(8, 0, 1),
    9: ApiVersionsSpec(9, 0, 1),
    10: ApiVersionsSpec(10, 0, 0),
    11: ApiVersionsSpec(11, 0, 0),
    12: ApiVersionsSpec(12, 0, 0),
    13: ApiVersionsSpec(13, 0, 0),
    14: ApiVersionsSpec(14, 0, 0),
    15: ApiVersionsSpec(15, 0, 0),
    16: ApiVersionsSpec(16, 0, 0)
}
API_VERSIONS_090 = {
    0: ApiVersionsSpec(0, 0, 0),
    1: ApiVersionsSpec(1, 0, 1),
    2: ApiVersionsSpec(2, 0, 0),
    3: ApiVersionsSpec(3, 0, 0),
    4: ApiVersionsSpec(4, 0, 0),
    5: ApiVersionsSpec(5, 0, 0),
    6: ApiVersionsSpec(6, 0, 0),
    7: ApiVersionsSpec(7, 0, 0),
    8: ApiVersionsSpec(8, 0, 1),
    9: ApiVersionsSpec(9, 0, 1),
    10: ApiVersionsSpec(10, 0, 0),
    11: ApiVersionsSpec(11, 0, 0),
    12: ApiVersionsSpec(12, 0, 0),
    13: ApiVersionsSpec(13, 0, 0),
    14: ApiVersionsSpec(14, 0, 0),
    15: ApiVersionsSpec(15, 0, 0),
    16: ApiVersionsSpec(16, 0, 0)
}



from __future__ import division
"""
Author: Emmett Butler, Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Producer"]
from collections import deque
import logging
import platform
import struct
import sys
import threading
import time
import weakref

from six import reraise

from .common import CompressionType
from .exceptions import (
    ERROR_CODES,
    KafkaException,
    InvalidMessageSize,
    MessageSizeTooLarge,
    NotLeaderForPartition,
    ProduceFailureError,
    ProducerQueueFullError,
    ProducerStoppedException,
    SocketDisconnectedError,
)
from .partitioners import RandomPartitioner
from .protocol import Message, ProduceRequest
from .utils.compat import iteritems, itervalues, Empty
from .utils.error_handlers import valid_int
from .utils import msg_protocol_version

log = logging.getLogger(__name__)


class Producer(object):
    """Implements asynchronous producer logic similar to the JVM driver.

    It creates a thread of execution for each broker that is the leader of
    one or more of its topic's partitions. Each of these threads (which may
    use `threading` or some other parallelism implementation like `gevent`)
    is associated with a queue that holds the messages that are waiting to be
    sent to that queue's broker.
    """
    def __init__(self,
                 cluster,
                 topic,
                 partitioner=None,
                 compression=CompressionType.NONE,
                 max_retries=3,
                 retry_backoff_ms=100,
                 required_acks=1,
                 ack_timeout_ms=10 * 1000,
                 max_queued_messages=100000,
                 min_queued_messages=70000,
                 linger_ms=5 * 1000,
                 # XXX 0 default here mirrors previous behavior - should default have a
                 # nonzero wait to save CPU cycles?
                 queue_empty_timeout_ms=0,
                 block_on_queue_full=True,
                 max_request_size=1000012,
                 sync=False,
                 delivery_reports=False,
                 pending_timeout_ms=5 * 1000,
                 auto_start=True,
                 serializer=None):
        """Instantiate a new AsyncProducer

        :param cluster: The cluster to which to connect
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param topic: The topic to which to produce messages
        :type topic: :class:`pykafka.topic.Topic`
        :param partitioner: The partitioner to use during message production
        :type partitioner: :class:`pykafka.partitioners.BasePartitioner`
        :param compression: The type of compression to use.
        :type compression: :class:`pykafka.common.CompressionType`
        :param max_retries: How many times to attempt to produce a given batch of
            messages before raising an error. Allowing retries will potentially change
            the ordering of records because if two records are sent to a single partition,
            and the first fails and is retried but the second succeeds, then the second
            record may appear first. If you want to completely disallow message
            reordering, use `sync=True`.
        :type max_retries: int
        :param retry_backoff_ms: The amount of time (in milliseconds) to
            back off during produce request retries. This does not equal the total time
            spent between message send attempts, since that number can be influenced
            by other kwargs, including `linger_ms` and `socket_timeout_ms`.
        :type retry_backoff_ms: int
        :param required_acks: The number of other brokers that must have
            committed the data to their log and acknowledged this to the leader
            before a request is considered complete
        :type required_acks: int
        :param ack_timeout_ms: The amount of time (in milliseconds) to wait for
            acknowledgment of a produce request on the server.
        :type ack_timeout_ms: int
        :param max_queued_messages: The maximum number of messages the producer
            can have waiting to be sent to the broker. If messages are sent
            faster than they can be delivered to the broker, the producer will
            either block or throw an exception based on the preference specified
            with block_on_queue_full.
        :type max_queued_messages: int
        :param min_queued_messages: The minimum number of messages the producer
            can have waiting in a queue before it flushes that queue to its
            broker (must be greater than 0). This paramater can be used to
            control the number of messages sent in one batch during async
            production. This parameter is automatically overridden to 1
            when `sync=True`.
        :type min_queued_messages: int
        :param linger_ms: This setting gives the upper bound on the delay for
            batching: once the producer gets min_queued_messages worth of
            messages for a broker, it will be sent immediately regardless of
            this setting.  However, if we have fewer than this many messages
            accumulated for this partition we will 'linger' for the specified
            time waiting for more records to show up. linger_ms=0 indicates no
            lingering - messages are sent as fast as possible after they are
            `produce()`d.
        :type linger_ms: int
        :param queue_empty_timeout_ms: The amount of time in milliseconds for which
            the producer's worker threads should block when no messages are available
            to flush to brokers. After each `linger_ms` interval, the worker thread
            checks for the presence of at least one message in its queue. If there is
            not at least one, it enters an "empty wait" period for
            `queue_empty_timeout_ms` before starting a new `linger_ms` wait loop. If
            `queue_empty_timeout_ms` is 0, this "empty wait" period is a noop, and
            flushes will continue to be attempted at intervals of `linger_ms`, even
            when the queue is empty. If `queue_empty_timeout_ms` is a positive integer,
            this "empty wait" period will last for at most that long, but it ends earlier
            if a message is `produce()`d before that time. If `queue_empty_timeout_ms` is
            -1, the "empty wait" period can only be stopped (and the worker thread killed)
            by a call to either `produce()` or `stop()` - it will never time out.
        :type queue_empty_timeout_ms: int
        :param block_on_queue_full: When the producer's message queue for a
            broker contains max_queued_messages, we must either stop accepting
            new messages (block) or throw an error. If True, this setting
            indicates we should block until space is available in the queue.
            If False, we should throw an error immediately.
        :type block_on_queue_full: bool
        :param max_request_size:
            The maximum size of a request in bytes. This is also effectively a
            cap on the maximum record size. Note that the server has its own
            cap on record size which may be different from this. This setting
            will limit the number of record batches the producer will send in a
            single request to avoid sending huge requests.
        :type max_request_size: int
        :param sync: Whether calls to `produce` should wait for the message to
            send before returning.  If `True`, an exception will be raised from
            `produce()` if delivery to kafka failed.
        :type sync: bool
        :param delivery_reports: If set to `True`, the producer will maintain a
            thread-local queue on which delivery reports are posted for each
            message produced.  These must regularly be retrieved through
            `get_delivery_report()`, which returns a 2-tuple of
            :class:`pykafka.protocol.Message` and either `None` (for success)
            or an `Exception` in case of failed delivery to kafka. If
            `get_delivery_report()` is not called regularly with this setting enabled,
            memory usage will grow unbounded. This setting is ignored when `sync=True`.
        :type delivery_reports: bool
        :param pending_timeout_ms: The amount of time (in milliseconds) to wait for
            delivery reports to be returned from the broker during a `produce()` call.
            Also, the time in ms to wait during a `stop()` call for all messages to be
            marked as delivered. -1 indicates that these calls should block indefinitely.
            Differs from `ack_timeout_ms` in that `ack_timeout_ms` is a value sent to the
            broker to control the broker-side timeout, while `pending_timeout_ms` is used
            internally by pykafka and not sent to the broker.
        :type pending_timeout_ms:
        :param auto_start: Whether the producer should begin communicating
            with kafka after __init__ is complete. If false, communication
            can be started with `start()`.
        :type auto_start: bool
        :param serializer: A function defining how to serialize messages to be sent
            to Kafka. A function with the signature d(value, partition_key) that
            returns a tuple of (serialized_value, serialized_partition_key). The
            arguments passed to this function are a message's value and partition key,
            and the returned data should be these fields transformed according to the
            client code's serialization logic.  See `pykafka.utils.__init__` for stock
            implemtations.
        :type serializer: function
        """
        self._cluster = cluster
        self._protocol_version = msg_protocol_version(cluster._broker_version)
        self._topic = topic
        self._partitioner = partitioner or RandomPartitioner()
        self._compression = compression
        if self._compression == CompressionType.SNAPPY and \
                platform.python_implementation == "PyPy":
            log.warning("Caution: python-snappy segfaults when attempting to compress "
                        "large messages under PyPy")
        self._max_retries = valid_int(max_retries, allow_zero=True)
        self._retry_backoff_ms = valid_int(retry_backoff_ms)
        self._required_acks = valid_int(required_acks, allow_zero=True,
                                        allow_negative=True)
        self._ack_timeout_ms = valid_int(ack_timeout_ms, allow_zero=True)
        self._max_queued_messages = valid_int(max_queued_messages, allow_zero=True)
        self._min_queued_messages = max(1, valid_int(min_queued_messages)
                                        if not sync else 1)
        self._linger_ms = valid_int(linger_ms, allow_zero=True)
        self._queue_empty_timeout_ms = valid_int(queue_empty_timeout_ms,
                                                 allow_zero=True, allow_negative=True)
        self._block_on_queue_full = block_on_queue_full
        self._max_request_size = valid_int(max_request_size)
        self._synchronous = sync
        self._worker_exception = None
        self._owned_brokers = None
        self._delivery_reports = (_DeliveryReportQueue(self._cluster.handler)
                                  if delivery_reports or self._synchronous
                                  else _DeliveryReportNone())
        self._pending_timeout_ms = pending_timeout_ms
        self._auto_start = auto_start
        self._serializer = serializer
        self._running = False
        self._update_lock = self._cluster.handler.Lock()
        if self._auto_start:
            self.start()

    def __del__(self):
        if log:  # in case log is finalized before self
            log.debug("Finalising {}".format(self))
        self.stop()

    def _raise_worker_exceptions(self):
        """Raises exceptions encountered on worker threads"""
        if self._worker_exception is not None:
            reraise(*self._worker_exception)

    def __repr__(self):
        return "<{module}.{name} at {id_}>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self))
        )

    def __enter__(self):
        """Context manager entry point - start the producer"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Context manager exit point - stop the producer"""
        self.stop()

    def start(self):
        """Set up data structures and start worker threads"""
        if not self._running:
            self._setup_owned_brokers()
            self._running = True
        self._raise_worker_exceptions()

    def _update(self):
        """Update the producer and cluster after an ERROR_CODE

        Also re-produces messages that were in queues at the time the update
        was triggered
        """
        # only allow one thread to be updating the producer at a time
        with self._update_lock:
            if self._owned_brokers is not None:
                for owned_broker in list(self._owned_brokers.values()):
                    owned_broker.stop()
            self._cluster.update()
            queued_messages = self._setup_owned_brokers()
            if len(queued_messages):
                log.debug("Re-producing %d queued messages after update",
                          len(queued_messages))
                for message in queued_messages:
                    self._produce(message)

    def _setup_owned_brokers(self):
        """Instantiate one OwnedBroker per broker

        If there are already OwnedBrokers instantiated, safely stop and flush them
        before creating new ones.
        """
        queued_messages = []
        if self._owned_brokers is not None:
            brokers = list(self._owned_brokers.keys())
            for broker in brokers:
                owned_broker = self._owned_brokers.pop(broker)
                owned_broker.stop()

                # loop because flush is not guaranteed to empty owned
                # broker queue
                while True:
                    batch = owned_broker.flush(self._linger_ms,
                                               self._max_request_size,
                                               release_pending=False,
                                               wait=False)
                    if not batch:
                        break
                    queued_messages.extend(batch)

        self._owned_brokers = {}
        for partition in self._topic.partitions.values():
            if partition.leader.id not in self._owned_brokers:
                self._owned_brokers[partition.leader.id] = OwnedBroker(
                    self, partition.leader)
        return queued_messages

    def stop(self):
        """Mark the producer as stopped, and wait until all messages to be sent"""
        def get_queue_readers():
            if not self._owned_brokers:
                return []
            return [owned_broker._queue_reader_worker
                    for owned_broker in self._owned_brokers.values()
                    if owned_broker.running]

        def stop_owned_brokers():
            self._wait_all()
            if self._owned_brokers is not None:
                for owned_broker in self._owned_brokers.values():
                    owned_broker.stop()

        while self._running:
            queue_readers = get_queue_readers()
            stop_owned_brokers()
            if len(queue_readers) == 0:
                self._running = False
            else:
                # The join() here works because new queue readers are spawned during the
                # execution of the old ones i.e: after a queue reader in its call to
                # producer._send_request encounters either a SocketDisconnectedError or
                # NotLeaderForPartition.ERROR_CODE it updates the cluster and sets up
                # owned_brokers, starting new queue reader threads. Because of this when
                # we call self.get_queue_readers, we get the newly created queue readers,
                # and so try to stop and join them etc ... until they all stop WITHOUT
                # encountering problems in producer._send_request.
                for queue_reader in queue_readers:
                    queue_reader.join()

    def _produce_has_timed_out(self, start_time):
        """Indicates whether enough time has passed since start_time for a `produce()`
            call to timeout
        """
        if self._pending_timeout_ms == -1:
            return False
        return time.time() * 1000 - start_time > self._pending_timeout_ms

    def produce(self, message, partition_key=None, timestamp=None):
        """Produce a message.

        :param message: The message to produce (use None to send null)
        :type message: bytes
        :param partition_key: The key to use when deciding which partition to send this
            message to. This key is passed to the `partitioner`, which may or may not
            use it in deciding the partition. The default `RandomPartitioner` does not
            use this key, but the optional `HashingPartitioner` does.
        :type partition_key: bytes
        :param timestamp: The timestamp at which the message is produced (requires
            broker_version >= 0.10.0)
        :type timestamp: `datetime.datetime`
        :return: The :class:`pykafka.protocol.Message` instance that was
            added to the internal message queue
        """
        if self._serializer is None:
            if partition_key is not None and type(partition_key) is not bytes:
                raise TypeError("Producer.produce accepts a bytes object as partition_key, "
                                "but it got '%s'", type(partition_key))
            if message is not None and type(message) is not bytes:
                raise TypeError("Producer.produce accepts a bytes object as message, but it "
                                "got '%s'", type(message))
        if timestamp is not None and self._protocol_version < 1:
            raise RuntimeError("Producer.produce got a timestamp with protocol 0")
        if not self._running:
            raise ProducerStoppedException()
        if self._serializer is not None:
            message, partition_key = self._serializer(message, partition_key)

        partitions = list(self._topic.partitions.values())
        partition_id = self._partitioner(partitions, partition_key).id

        msg = Message(value=message,
                      partition_key=partition_key,
                      partition_id=partition_id,
                      timestamp=timestamp,
                      protocol_version=self._protocol_version,
                      # We must pass our thread-local Queue instance directly,
                      # as results will be written to it in a worker thread
                      delivery_report_q=self._delivery_reports.queue)
        self._produce(msg)

        if self._synchronous:
            req_time = time.time() * 1000
            reported_msg = None
            while not self._produce_has_timed_out(req_time):
                self._raise_worker_exceptions()
                self._cluster.handler.sleep()
                try:
                    reported_msg, exc = self.get_delivery_report(timeout=1)
                    break
                except Empty:
                    continue
                except ValueError:
                    raise ProduceFailureError("Error retrieving delivery report")
            if reported_msg is not msg:
                raise ProduceFailureError("Delivery report not received after timeout")
            if exc is not None:
                raise exc
        self._raise_worker_exceptions()
        return msg

    def get_delivery_report(self, block=True, timeout=None):
        """Fetch delivery reports for messages produced on the current thread

        Returns 2-tuples of a `pykafka.protocol.Message` and either `None`
        (for successful deliveries) or `Exception` (for failed deliveries).
        This interface is only available if you enabled `delivery_reports` on
        init (and you did not use `sync=True`)

        :param block: Whether to block on dequeueing a delivery report
        :type block: bool
        :param timeout: How long (in seconds) to block before returning None
        ;type timeout: int
        """
        try:
            return self._delivery_reports.queue.get(block, timeout)
        except AttributeError:
            raise KafkaException("Delivery-reporting is disabled")

    def _produce(self, message):
        """Enqueue a message for the relevant broker
        Attempts to update metadata in response to missing brokers.
        :param message: Message with valid `partition_id`, ready to be sent
        :type message: `pykafka.protocol.Message`
        """
        success = False
        retry = 0
        while not success:
            leader_id = self._topic.partitions[message.partition_id].leader.id
            if leader_id in self._owned_brokers:
                self._owned_brokers[leader_id].enqueue(message)
                success = True
            else:
                retry += 1
                if retry < 10:
                    log.debug("Failed to enqueue produced message. Updating metdata.")
                    self._update()
                else:
                    raise ProduceFailureError("Message could not be enqueued due to missing broker "
                                              "metadata for broker {}".format(leader_id))
                success = False

    def _mark_as_delivered(self, owned_broker, message_batch, req):
        owned_broker.increment_messages_pending(-1 * len(message_batch))
        req.delivered += len(message_batch)
        for msg in message_batch:
            self._delivery_reports.put(msg)

    def _send_request(self, message_batch, owned_broker):
        """Send the produce request to the broker and handle the response.

        :param message_batch: An iterable of messages to send
        :type message_batch: iterable of `pykafka.protocol.Message`
        :param owned_broker: The broker to which to send the request
        :type owned_broker: :class:`pykafka.producer.OwnedBroker`
        """
        req = ProduceRequest(
            compression_type=self._compression,
            required_acks=self._required_acks,
            timeout=self._ack_timeout_ms,
            broker_version=self._cluster._broker_version
        )
        req.delivered = 0
        for msg in message_batch:
            req.add_message(msg, self._topic.name, msg.partition_id)
        log.debug("Sending %d messages to broker %d",
                  len(message_batch), owned_broker.broker.id)

        def _get_partition_msgs(partition_id, req):
            """Get all the messages for the partitions from the request."""
            return (
                mset
                for topic, partitions in iteritems(req.msets)
                for p_id, mset in iteritems(partitions)
                if p_id == partition_id
            )

        try:
            response = owned_broker.broker.produce_messages(req)
            if self._required_acks == 0:  # and thus, `response` is None
                self._mark_as_delivered(owned_broker, message_batch, req)
                return

            # Kafka either atomically appends or rejects whole MessageSets, so
            # we define a list of potential retries thus:
            to_retry = []  # (MessageSet, Exception) tuples

            for topic, partitions in iteritems(response.topics):
                for partition, presponse in iteritems(partitions):
                    if presponse.err == 0:
                        messages = req.msets[topic][partition].messages
                        for i, message in enumerate(messages):
                            message.offset = presponse.offset + i
                        self._mark_as_delivered(owned_broker, messages, req)
                        continue  # All's well
                    if presponse.err == NotLeaderForPartition.ERROR_CODE:
                        # Update cluster metadata to get new leader
                        self._update()
                    info = "Produce request for {}/{} to {}:{} failed with error code {}.".format(
                        topic,
                        partition,
                        owned_broker.broker.host,
                        owned_broker.broker.port,
                        presponse.err)
                    log.warning(info)
                    exc = ERROR_CODES[presponse.err](info)
                    to_retry.extend(
                        (mset, exc)
                        for mset in _get_partition_msgs(partition, req))
        except (SocketDisconnectedError, struct.error) as exc:
            log.warning('Error encountered when producing to broker %s:%s. Retrying.',
                        owned_broker.broker.host,
                        owned_broker.broker.port)
            self._update()
            to_retry = [
                (mset, exc)
                for topic, partitions in iteritems(req.msets)
                for p_id, mset in iteritems(partitions)
            ]

        log.debug("Successfully sent {}/{} messages to broker {}".format(
            req.delivered, len(message_batch), owned_broker.broker.id))

        if to_retry:
            self._cluster.handler.sleep(self._retry_backoff_ms / 1000)
            owned_broker.increment_messages_pending(-1 * len(to_retry))
            for mset, exc in to_retry:
                # XXX arguably, we should try to check these non_recoverables
                # for individual messages in _produce and raise errors there
                # right away, rather than failing a whole batch here?
                non_recoverable = type(exc) in (InvalidMessageSize,
                                                MessageSizeTooLarge)
                for msg in mset.messages:
                    if (non_recoverable or msg.produce_attempt >= self._max_retries):
                        self._delivery_reports.put(msg, exc)
                        log.error("Message not delivered!! %r" % exc)
                    else:
                        msg.produce_attempt += 1
                        self._produce(msg)

    def _wait_all(self):
        """Block until all pending messages are sent or until pending_timeout_ms

        "Pending" messages are those that have been used in calls to `produce`
        and have not yet been acknowledged in a response from the broker
        """
        log.info("Blocking until all messages are sent or until pending_timeout_ms")
        start_time = time.time() * 1000
        while any(q.message_is_pending() for q in itervalues(self._owned_brokers)) and \
                not self._produce_has_timed_out(start_time):
            self._cluster.handler.sleep(.3)
            self._raise_worker_exceptions()


class OwnedBroker(object):
    """An abstraction over a broker connected to by the producer

    An OwnedBroker object contains thread-synchronization primitives
    and message queue corresponding to a single broker for this producer.

    :ivar lock: The lock used to control access to shared resources for this
        queue
    :ivar flush_ready: A condition variable that indicates that the queue is
        ready to be flushed via requests to the broker
    :ivar slot_available: A condition variable that indicates that there is
        at least one position free in the queue for a new message
    :ivar queue: The message queue for this broker. Contains messages that have
        been supplied as arguments to `produce()` waiting to be sent to the
        broker.
    :type queue: collections.deque
    :ivar messages_pending: A counter indicating how many messages have been
        enqueued for this broker and not yet sent in a request.
    :type messages_pending: int
    :ivar producer: The producer to which this OwnedBroker instance belongs
    :type producer: :class:`pykafka.producer.AsyncProducer`
    :param auto_start: Whether the OwnedBroker should start flushing all
        waiting messages and send to kafka after __init__ is complete. If
        false, communication can be started with `start()`.
    :type auto_start: bool
    """
    def __init__(self, producer, broker, auto_start=True):
        self.producer = weakref.proxy(producer)
        self.broker = broker
        self.lock = self.producer._cluster.handler.RLock()
        self.flush_ready = self.producer._cluster.handler.Event()
        self.has_message = self.producer._cluster.handler.Event()
        self.slot_available = self.producer._cluster.handler.Event()
        self.queue = deque()
        self.messages_pending = 0
        self.running = True
        self._auto_start = auto_start

        if self._auto_start:
            self.start()

    def cleanup(self):
        if not self.slot_available.is_set():
            self.slot_available.set()

    def start(self):
        def queue_reader():
            while self.running:
                try:
                    batch = self.flush(self.producer._linger_ms, self.producer._max_request_size)
                    if batch:
                        self.producer._send_request(batch, self)
                except Exception:
                    # surface all exceptions to the main thread
                    self.producer._worker_exception = sys.exc_info()
                    break
            self.cleanup()
            log.info("Worker exited for broker %s:%s", self.broker.host,
                     self.broker.port)
        log.info("Starting new produce worker for broker %s", self.broker.id)
        name = "pykafka.OwnedBroker.queue_reader for broker {}".format(self.broker.id)
        self._queue_reader_worker = self.producer._cluster.handler.spawn(
            queue_reader, name=name)

    def stop(self):
        # explicitly set has_message to kill any infinite waits triggered by
        # queue_empty_timeout_ms=-1
        with self.lock:
            if not self.has_message.is_set():
                self.has_message.set()
        self.running = False

    def increment_messages_pending(self, amnt):
        with self.lock:
            self.messages_pending += amnt
            self.messages_pending = max(0, self.messages_pending)

    def message_is_pending(self):
        """
        Indicates whether there are currently any messages that have been
            `produce()`d and not yet sent to the broker
        """
        return self.messages_pending > 0

    def enqueue(self, message):
        """Push message onto the queue

        :param message: The message to push onto the queue
        :type message: `pykafka.protocol.Message`
        """
        self._wait_for_slot_available()
        with self.lock:
            self.queue.appendleft(message)
            self.increment_messages_pending(1)
            if len(self.queue) >= self.producer._min_queued_messages:
                if not self.flush_ready.is_set():
                    self.flush_ready.set()
            if not self.has_message.is_set():
                self.has_message.set()

    def flush(self, linger_ms, max_request_size, release_pending=False, wait=True):
        """Pop messages from the end of the queue

        :param linger_ms: How long (in milliseconds) to wait for the queue
            to contain messages before flushing
        :type linger_ms: int
        :param max_request_size: The max size should each batch of messages
            should be in bytes
        :type max_request_size: int
        :param release_pending: Whether to decrement the messages_pending
            counter when the queue is flushed. True means that the messages
            popped from the queue will be discarded unless re-enqueued
            by the caller.
        :type release_pending: bool
        :param wait: If True, wait for the event indicating a flush is ready. If False,
            attempt a flush immediately without waiting
        :type wait: bool
        """
        # Q: why not simply wait for flush_ready here? do we need a separate Event for
        #    has_message?
        # A: If we're blocking on flush_ready with an empty queue, a single event arriving
        #    does not mean we're ready to flush. We could flush whenever the
        #    current linger_ms interval ends, but the better way is to pause the linger
        #    loop when the queue is empty, restarting it when a message is added. Doing
        #    this without two Events would require _wait_for_flush_ready to be modal,
        #    returning a value indicating whether a flush should happen or whether it
        #    returned for the sole purpose of unblocking the linger loop. This design is
        #    cleaner.
        self._wait_for_has_message(self.producer._queue_empty_timeout_ms)
        if wait:
            self._wait_for_flush_ready(linger_ms)
        with self.lock:
            batch = []
            batch_size_in_bytes = 0
            while len(self.queue) > 0:
                if not self.running:
                    return []
                peeked_message = self.queue[-1]

                if peeked_message and peeked_message.value is not None:
                    if len(peeked_message) > max_request_size:
                        exc = MessageSizeTooLarge(
                            "Message size larger than max_request_size: {}".format(max_request_size)
                        )
                        log.warning(exc)
                        # bind the MessageSizeTooLarge error the delivery
                        # report and remove it from the producer queue
                        message = self.queue.pop()
                        # don't use producer.delivery_report_q here to enable
                        # integration tests that test the OwnedBroker without a
                        # Producer
                        if peeked_message.delivery_report_q is not None:
                            peeked_message.delivery_report_q.put((message, exc))
                        # remove from pending message count
                        self.increment_messages_pending(-1)
                        continue

                    # test if adding the message would go over the
                    # max_request_size. if it would, break out of loop
                    elif batch_size_in_bytes + len(peeked_message) > max_request_size:
                        log.debug("max_request_size reached. producing batch")
                        # if we did not fully empty the queue. reset the
                        # flush_ready so we send another batch immediately
                        self.flush_ready.set()
                        break

                message = self.queue.pop()
                batch_size_in_bytes += len(message)
                batch.append(message)

            if release_pending:
                self.increment_messages_pending(-1 * len(batch))
            if not self.slot_available.is_set():
                self.slot_available.set()
        if not self.running:
            return []
        return batch

    def _wait_for_flush_ready(self, linger_ms):
        """Block until the queue is ready to be flushed

        If the queue does not contain at least one message after blocking for
        `linger_ms` milliseconds, return.

        :param linger_ms: How long (in milliseconds) to wait for the queue
            to contain messages before returning
        :type linger_ms: int
        """
        if len(self.queue) < self.producer._min_queued_messages:
            with self.lock:
                if len(self.queue) < self.producer._min_queued_messages:
                    self.flush_ready.clear()
            if linger_ms > 0:
                self.flush_ready.wait((linger_ms / 1000))

    def _wait_for_has_message(self, timeout_ms):
        """Block until the queue has at least one slot containing a message

        :param timeout_ms: The amount of time in milliseconds to wait for a message
            to be enqueued. -1 indicates infinite waiting; in this case a thread waiting
            on this call can only be killed by a call to `stop()`.
        :type timeout_ms: int
        """
        if len(self.queue) == 0 and self.running:
            with self.lock:
                if len(self.queue) == 0 and self.running:
                    self.has_message.clear()
            if timeout_ms != -1:
                self.has_message.wait(timeout_ms / 1000)
            else:
                # infinite wait that doesn't break under gevent
                while not self.has_message.is_set():
                    self.producer._cluster.handler.sleep()
                    self.has_message.wait(5)

    def _wait_for_slot_available(self):
        """Block until the queue has at least one slot not containing a message"""
        if len(self.queue) >= self.producer._max_queued_messages:
            with self.lock:
                if len(self.queue) >= self.producer._max_queued_messages:
                    self.slot_available.clear()
            if self.producer._block_on_queue_full:
                while not self.slot_available.is_set():
                    self.producer._cluster.handler.sleep()
                    self.slot_available.wait(5)
            else:
                raise ProducerQueueFullError("Queue full for broker %d",
                                             self.broker.id)


class _DeliveryReportQueue(threading.local):
    """Helper that instantiates a new report queue on every calling thread"""
    def __init__(self, handler):
        self.queue = handler.Queue()

    @staticmethod
    def put(msg, exc=None):
        msg.delivery_report_q.put((msg, exc))


class _DeliveryReportNone(object):
    """Stand-in for when _DeliveryReportQueue has been disabled"""
    def __init__(self):
        self.queue = None

    @staticmethod
    def put(msg, exc=None):
        return
